import strax
import numba
import numpy as np

export, __all__ = strax.exporter()


# (5-10x) faster than np.sort(order=...), as np.sort looks at all fields
# TODO: maybe this should be a factory?
@export
@numba.jit(nopython=True, nogil=True, cache=True)
def sort_by_time(x):
    """Sort pulses by time, then channel.

    Assumes you have no more than 10k channels, and records don't span
    more than 100 days. TODO: FIX this
    """
    # I couldn't get fast argsort on multiple keys to work in numba
    # So, let's make a single key...
    sort_key = (x['time'] - x['time'].min()) * 10000 + x['channel']
    sort_i = np.argsort(sort_key)
    return x[sort_i]


@export
@numba.jit(nopython=True, nogil=True, cache=True)
def first_index_not_below(arr, t):
    """Return first index of array >= t, or len(arr) if no such found"""
    for i, x in enumerate(arr):
        if x >= t:
            return i
    return len(arr)


@export
def endtime(x):
    """Return endtime of intervals x"""
    if 'endtime' in x.dtype.names:
        return x['endtime']
    return x['time'] + x['length'] * x['dt']


@export
# TODO: somehow numba compilation hangs on this one? reproduce / file issue?
# numba.jit(nopython=True, nogil=True, cache=True)
def from_break(x, safe_break=10000, left=True, tolerant=False):
    """Return records on side of a break at least safe_break long
    If there is no such break, return the best break found.
    """
    # TODO: This is extremely rough. Better to find proper gaps, and if we
    # know the timing of the readers, consider breaks at end and start too.
    break_i = find_break_i(x, safe_break=safe_break, tolerant=tolerant)

    if left:
        return x[:break_i]
    else:
        return x[break_i:]


@export
class NoBreakFound(Exception):
    pass


@export
@numba.jit(nopython=True, nogil=True, cache=True)
def find_break_i(x, safe_break, tolerant=True):
    """Returns LAST index of x whose time is more than safe_break away
    from the x before
    :param tolerant: if no break found, yield an as good as possible break
    anyway.
    """
    max_gap = 0
    max_gap_i = -1
    for _i in range(len(x) - 1):
        i = len(x) - 1 - _i
        gap = x[i]['time'] - x[i - 1]['time']
        if gap >= safe_break:
            return i
        if gap > max_gap:
            max_gap_i = i
            max_gap = gap

    if not tolerant:
        raise NoBreakFound

    print("\t\tDid not find safe break, using largest available break: ",
          max_gap,
          " ns")
    return max_gap_i


@export
def fully_contained_in(things, containers):
    """Return array of len(things) with index of interval in containers
    for which things are fully contained in a container, or -1 if no such
    exists. We assume all intervals are sorted by time, and b_intervals
    nonoverlapping.
    """
    result = np.ones(len(things), dtype=np.int32) * -1
    a_starts = things['time']
    b_starts = containers['time']
    a_ends = strax.endtime(things)
    b_ends = strax.endtime(containers)
    _fc_in(a_starts, b_starts, a_ends, b_ends, result)
    return result


@numba.jit(nopython=True, nogil=True, cache=True)
def _fc_in(a_starts, b_starts, a_ends, b_ends, result):
    b_i = 0
    for a_i in range(len(a_starts)):
        # Skip ahead one or more b's if we're beyond them
        # Note <= in second condition: end is an exclusive bound
        while b_i < len(b_starts) and b_ends[b_i] <= a_starts[a_i]:
            b_i += 1
        if b_i == len(b_starts):
            break

        # Check for containment. We only need to check one b, since bs
        # are nonoverlapping
        if b_starts[b_i] <= a_starts[a_i] and a_ends[a_i] <= b_ends[b_i]:
            result[a_i] = b_i


@export
def split_by_containment(things, containers):
    """Return list of thing-arrays contained in each container

    Assumes everything is sorted, and containers are nonoverlapping
    """
    if not len(containers):
        return []

    which_container = fully_contained_in(things, containers)

    mask = which_container != -1
    things = things[mask]
    which_container = which_container[mask]
    things_split = np.split(
        things,
        np.where(np.diff(which_container))[0] + 1)

    # Insert empties for containers with nothing
    for c in np.setdiff1d(np.arange(len(containers)),
                          np.unique(which_container)):
        if c == 0:
            continue   # np.split already produces an empty in this case?
        things_split.insert(c, things[:0])

    return things_split



from concurrent.futures import Future, TimeoutError
import heapq
import threading
import typing
import logging

from strax.utils import exporter
export, __all__ = exporter()


@export
class MailboxException(Exception):
    pass


@export
class MailboxReadTimeout(MailboxException):
    pass


@export
class MailboxFullTimeout(MailboxException):
    pass


@export
class InvalidMessageNumber(MailboxException):
    pass


@export
class MailBoxAlreadyClosed(MailboxException):
    pass


@export
class MailboxKilled(MailboxException):
    pass


@export
class Mailbox:
    """Publish/subscribe mailbox for builing complex pipelines
    out of simple iterators, using multithreading.

    A sender can be any iterable. To read from the mailbox, either:
     1. Use .subscribe() to get an iterator.
        You can only use one of these per thread.
     2. Use .add_subscriber(f) to subscribe the function f.
        f should take an iterator as its first argument
        (and actually iterate over it, of course).

    Each sender and receiver is wrapped in a thread, so they can be paused:
     - senders, if the mailbox is full;
     - readers, if they call next() but the next message is not yet available.

    Any futures sent in are awaited before they are passed to receivers.

    Exceptions in a sender cause MailboxKilled to be raised in each reader.
    If the reader doesn't catch this, and it writes to another mailbox,
    this therefore kills that mailbox (raises MailboxKilled for each reader)
    as well. Thus MailboxKilled exceptions travel downstream in pipelines.

    Sender threads are not killed by exceptions raise in readers.
    To kill sender threads too, use .kill(upstream=True). Even this does not
    propagate further upstream than the immediate sender threads.
    """

    def __init__(self,
                 name='mailbox',
                 timeout=120,
                 max_messages=30):
        self.name = name
        self.timeout = timeout
        self.max_messages = max_messages

        self.closed = False
        self.force_killed = False
        self.killed = False
        self.killed_because = None

        self._mailbox = []
        self._subscribers_have_read = []
        self._n_sent = 0
        self._threads = []
        self._lock = threading.RLock()
        self._read_condition = threading.Condition(lock=self._lock)
        self._write_condition = threading.Condition(lock=self._lock)

        self.log = logging.getLogger(self.name)
        self.log.debug("Initialized")

    def add_sender(self, source, name=None):
        """Configure mailbox to read from an iterable source

        :param source: Iterable to read from
        :param name: Name of the thread in which the function will run.
        Defaults to source:<mailbox_name>
        """
        if name is None:
            name = f'source:{self.name}'
        t = threading.Thread(target=self._send_from,
                             name=name,
                             args=(source,))
        self._threads.append(t)

    def add_reader(self, subscriber, name=None, **kwargs):
        """Subscribe a function to the mailbox.

        Function should accept the generator over messages as first argument.
        kwargs will be passed to function.

        :param name: Name of the thread in which the function will run.
        Defaults to read_<number>:<mailbox_name>
        """
        if name is None:
            name = f'read_{self._n_subscribers}:{self.name}'
        t = threading.Thread(target=subscriber,
                             name=name,
                             args=(self.subscribe(),),
                             kwargs=kwargs)
        self._threads.append(t)

    def subscribe(self):
        """Return generator over messages in the mailbox
        """
        with self._lock:
            subscriber_i = self._n_subscribers
            self._subscribers_have_read.append(-1)
            self.log.debug(f"Added subscriber {subscriber_i}")
            return self._read(subscriber_i=subscriber_i)

    def start(self):
        for t in self._threads:
            t.start()

    def kill(self, upstream=True, reason=None):
        with self._lock:
            if upstream:
                self.force_killed = True
            self.killed = True
            self.killed_because = reason
            self._read_condition.notify_all()

    def cleanup(self):
        for t in self._threads:
            t.join(timeout=self.timeout)
            if t.is_alive():
                raise RuntimeError("Thread %s did not terminate!" % t.name)

    def _send_from(self, iterable):
        """Send to mailbox from iterable, exiting appropriately if an
        exception is thrown
        """
        try:
            for x in iterable:
                self.send(x)
        except MailboxKilled as e:
            # The iterable was reading from a mailbox, which has been killed.
            # Kill this mailbox too.
            self.kill(reason=e.args[0])
            # Do NOT raise! One traceback on the screen is enough.
        except Exception as e:
            try:
                e_args = e.args[0]
            except IndexError:
                e_args = ''
            self.kill(reason=f"{e.__class__.__name__}('{e_args}') "
                             f'in {threading.current_thread().name}')
            raise
        else:
            self.close()

    def send(self, msg, msg_number: typing.Union[int, None] = None):
        """Send a message.

        If the message is a future, receivers will be passed its result.
        (possibly waiting for completion if needed)

        If the mailbox is currently full, sleep until there
        is room for your message (or timeout occurs)
        """
        with self._lock:
            if self.closed:
                raise MailBoxAlreadyClosed(f"Can't send to closed {self.name}")
            if self.force_killed:
                self.log.debug(f"Sender found {self.name} force-killed")
                raise MailboxKilled(self.killed_because)
            if self.killed:
                self.log.debug("Send to killed mailbox: message lost")
                return

            # We accept int numbers or anything which equals to it's int(...)
            # (like numpy integers)
            if msg_number is None:
                msg_number = self._n_sent
            try:
                int(msg_number)
                assert msg_number == int(msg_number)
            except (ValueError, AssertionError):
                raise InvalidMessageNumber("Msg numbers must be integers")

            read_until = min(self._subscribers_have_read, default=-1)
            if msg_number <= read_until:
                raise InvalidMessageNumber(
                    f'Attempt to send message {msg_number} while '
                    f'subscribers already read {read_until}.')

            def can_write():
                return len(self._mailbox) < self.max_messages
            if not can_write():
                self.log.debug(self._subscribers_have_read)
                self.log.debug(f"Mailbox full, wait to send {msg_number}")
            if not self._write_condition.wait_for(can_write,
                                                  timeout=self.timeout):
                raise MailboxFullTimeout(f"{self.name} emptied too slow")

            heapq.heappush(self._mailbox, (msg_number, msg))
            self.log.debug(f"Sent {msg_number}")
            self._n_sent += 1
            self._read_condition.notify_all()

    def close(self):
        with self._lock:
            self.send(StopIteration)
            self.closed = True
        self.log.debug(f"Closed to incoming messages")

    def _read(self, subscriber_i):
        """Iterate over incoming messages in order.

        Your thread will sleep until the next message is available, or timeout
        expires (in which case MailboxReadTimeout is raised)
        """
        self.log.debug("Start reading")
        next_number = 0
        last_message = False

        while not last_message:
            with self._lock:

                # Wait for new messages
                def next_ready():
                    return self._has_msg(next_number) or self.killed
                if not next_ready():
                    self.log.debug(f"Checking/waiting for {next_number}")
                if not self._read_condition.wait_for(next_ready,
                                                     self.timeout):
                    raise MailboxReadTimeout(
                        f"{self.name} did not get {next_number} in time")

                if self.killed:
                    self.log.debug(f"Reader finds {self.name} killed")
                    raise MailboxKilled(self.killed_because)

                # Grab all messages we can yield
                to_yield = []
                while self._has_msg(next_number):
                    msg = self._get_msg(next_number)
                    if msg is StopIteration:
                        self.log.debug(f"{next_number} is StopIteration")
                        last_message = True
                    to_yield.append((next_number, msg))
                    next_number += 1

                if len(to_yield) > 1:
                    self.log.debug(f"Read {to_yield[0][0]}-{to_yield[-1][0]}")
                else:
                    self.log.debug(f"Read {to_yield[0][0]}")

                self._subscribers_have_read[subscriber_i] = next_number - 1

                # Clean up the mailbox
                while (len(self._mailbox)
                       and (min(self._subscribers_have_read)
                            >= self._lowest_msg_number)):
                    heapq.heappop(self._mailbox)
                self._write_condition.notify_all()

            for msg_number, msg in to_yield:
                if msg is StopIteration:
                    return
                elif isinstance(msg, Future):
                    if not msg.done():
                        self.log.debug(f"Waiting for future {msg_number}")
                        try:
                            res = msg.result(timeout=self.timeout)
                        except TimeoutError:
                            raise TimeoutError(
                                f"Future {msg_number} timed out!")
                        self.log.debug(f"Future {msg_number} completed")
                    else:
                        res = msg.result()
                        self.log.debug(f"Future {msg_number} was already done")
                    yield res
                else:
                    yield msg

        self.log.debug("Done reading")

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name}>"

    def _get_msg(self, number):
        for msg_number, msg in self._mailbox:
            if msg_number == number:
                return msg
        else:
            raise RuntimeError(f"Could not find message {number}")

    def _has_msg(self, number):
        """Return if mailbox has message number.

        Also returns True if mailbox is killed, so be sure to check
        self.killed after this!
        """
        if self.killed:
            return True
        return any([msg_number == number
                    for msg_number, _ in self._mailbox])

    @property
    def _n_subscribers(self):
        return len(self._subscribers_have_read)

    @property
    def _lowest_msg_number(self):
        return self._mailbox[0][0]



"""Plugin system for strax

A 'plugin' is something that outputs an array and gets arrays
from one or more other plugins.
"""
from concurrent.futures import wait
from enum import IntEnum
import itertools
from functools import partial
import typing

import numpy as np

import strax
export, __all__ = strax.exporter()


@export
class SaveWhen(IntEnum):
    """Plugin's preference for having it's data saved"""
    NEVER = 0         # Throw an error if the user lists it
    EXPLICIT = 1      # Save ONLY if the user lists it explicitly
    TARGET = 2        # Save if the user asks for it as a final target
    ALWAYS = 3        # Save even if the user does not list it


@export
class Plugin:
    """Plugin containing strax computation

    You should NOT instantiate plugins directly.
    Do NOT add unpickleable things (e.g. loggers) as attributes.
    """
    __version__ = '0.0.0'
    data_kind: str
    depends_on: tuple
    provides: str

    compressor = 'blosc'

    rechunk_on_save = True    # Saver is allowed to rechunk
    rechunk_input = False

    save_when = SaveWhen.ALWAYS
    parallel = False    # If True, compute() work is submitted to pool
    save_meta_only = False

    # These are set on plugin initialization, which is done in the core
    run_id: str
    config: typing.Dict
    deps: typing.List       # Dictionary of dependency plugin instances
    takes_config = dict()       # Config options
    compute_takes_chunk_i = False    # Autoinferred, no need to set yourself

    def __init__(self):
        if not hasattr(self, 'depends_on'):
            raise ValueError('depends_on not provided for '
                             f'{self.__class__.__name__}')
        self.post_compute = []
        self.on_close = []
        if self.rechunk_input and self.parallel:
            raise RuntimeError("Plugins that rechunk their input "
                               "cannot be parallelized")

    def startup(self):
        """Hook if plugin wants to do something after initialization."""
        pass

    def infer_dtype(self):
        """Return dtype of computed data;
        used only if no dtype attribute defined"""
        raise NotImplementedError

    def version(self, run_id=None):
        """Return version number applicable to the run_id.
        Most plugins just have a single version (in .__version__)
        but some may be at different versions for different runs
        (e.g. time-dependent corrections).
        """
        return self.__version__

    def metadata(self, run_id):
        """Metadata to save along with produced data"""
        return dict(
            run_id=run_id,
            data_type=self.provides,
            data_kind=self.data_kind,
            dtype=self.dtype,
            compressor=self.compressor,
            lineage=self.lineage)

    def dependencies_by_kind(self, require_time=None):
        """Return dependencies grouped by data kind
        i.e. {kind1: [dep0, dep1], kind2: [dep, dep]}
        :param require_time: If True, one dependency of each kind
        must provide time information. It will be put first in the list.

        If require_time is omitted, we will require time only if there is
        more than one data kind in the dependencies.
        """
        if require_time is None:
            require_time = \
                len(self.dependencies_by_kind(require_time=False)) > 1

        deps_by_kind = dict()
        key_deps = []
        for d in self.depends_on:
            k = self.deps[d].data_kind
            deps_by_kind.setdefault(k, [])

            # If this has time information, put it first in the list
            if (require_time
                    and 'time' in self.deps[d].dtype.names):
                key_deps.append(d)
                deps_by_kind[k].insert(0, d)
            else:
                deps_by_kind[k].append(d)

        if require_time:
            for k, d in deps_by_kind.items():
                if not d[0] in key_deps:
                    raise ValueError(f"For {self.__class__.__name__}, no "
                                     f"dependency of data kind {k} "
                                     "has time information!")

        return deps_by_kind

    def iter(self, iters, executor=None):
        """Iterate over dependencies and yield results
        :param iters: dict with iterators over dependencies
        :param executor: Executor to punt computation tasks to.
            If None, will compute inside the plugin's thread.
        """
        deps_by_kind = self.dependencies_by_kind()

        if len(deps_by_kind) > 1:
            # Sync the iterators that provide time info for each data kind
            # (first in deps_by_kind lists) by endtime
            iters.update(strax.sync_iters(
                partial(strax.same_stop, func=strax.endtime),
                {d[0]: iters[d[0]]
                 for d in deps_by_kind.values()}))

        # Convert to iterators over merged data for each kind
        new_iters = dict()
        for kind, deps in deps_by_kind.items():
            if len(deps) > 1:
                synced_iters = strax.sync_iters(
                    strax.same_length,
                    {d: iters[d] for d in deps})
                new_iters[kind] = strax.merge_iters(synced_iters.values())
            else:
                new_iters[kind] = iters[deps[0]]
        iters = new_iters

        if self.rechunk_input:
            iters = self.rechunk_input(iters)

        pending = []
        for chunk_i in itertools.count():
            try:
                if not self.check_next_ready_or_done(chunk_i):
                    # TODO: avoid duplication
                    # but also remain picklable...
                    self.close(wait_for=tuple(pending))
                    return
                compute_kwargs = {k: next(iters[k])
                                  for k in deps_by_kind}
            except StopIteration:
                self.close(wait_for=tuple(pending))
                return
            except Exception:
                self.close(wait_for=tuple(pending))
                raise

            if self.parallel and executor is not None:
                new_f = executor.submit(self.do_compute,
                                        chunk_i=chunk_i,
                                        **compute_kwargs)
                pending = [f for f in pending + [new_f]
                           if not f.done()]
                yield new_f
            else:
                yield self.do_compute(chunk_i=chunk_i, **compute_kwargs)

    def do_compute(self, *args, chunk_i=None, **kwargs):
        for i, x in enumerate(args):
            kwargs[self.depends_on[i]] = x

        if self.compute_takes_chunk_i:
            result = self.compute(**kwargs, chunk_i=chunk_i)
        else:
            result = self.compute(**kwargs)

        if isinstance(result, dict):
            if not len(result):
                # TODO: alt way of getting length?
                raise RuntimeError("if returning dict, must have a key")
            some_key = list(result.keys())[0]
            n = len(result[some_key])
            r = np.zeros(n, dtype=self.dtype)
            for k, v in result.items():
                r[k] = v
            result = r

        for p in self.post_compute:
            r = p(result, chunk_i=chunk_i)
            if r is not None:
                result = r

        return result

    def check_next_ready_or_done(self, chunk_i):
        return True

    def close(self, wait_for=tuple(), timeout=120):
        done, not_done = wait(wait_for, timeout=timeout)
        if len(not_done):
            raise RuntimeError(
                f"{len(not_done)} futures of {self.__class__.__name__}"
                "did not complete in time!")
        for x in self.on_close:
            x()

    def compute(self, **kwargs):
        raise NotImplementedError

##
# Special plugins
##


@export
class LoopPlugin(Plugin):
    """Plugin that disguises multi-kind data-iteration by an event loop
    """
    def compute(self, **kwargs):
        # If not otherwise specified, data kind to loop over
        # is that of the first dependency (e.g. events)
        # Can't be in __init__: deps not initialized then
        if hasattr(self, 'loop_over'):
            loop_over = self.loop_over
        else:
            loop_over = self.deps[self.depends_on[0]].data_kind

        # Group into lists of things (e.g. peaks)
        # contained in the base things (e.g. events)
        base = kwargs[loop_over]
        if len(base) > 1:
            assert np.all(base[1:]['time'] >= strax.endtime(base[:-1])), \
                f'{base}s overlap'

        for k, things in kwargs.items():
            if len(things) > 1:
                assert np.diff(things['time']).min() > 0, f'{k} not sorted'
            if k != loop_over:
                r = strax.split_by_containment(things, base)
                if len(r) != len(base):
                    raise RuntimeError(f"Split {k} into {len(r)}, "
                                       f"should be {len(base)}!")
                kwargs[k] = r

        results = np.zeros(len(base), dtype=self.dtype)
        for i in range(len(base)):
            r = self.compute_loop(base[i],
                                  **{k: kwargs[k][i]
                                     for k in self.dependencies_by_kind()
                                     if k != loop_over})

            # Convert from dict to array row:
            for k, v in r.items():
                results[i][k] = v

        return results

    def compute_loop(self, *args, **kwargs):
        raise NotImplementedError


@export
class MergeOnlyPlugin(Plugin):
    """Plugin that merges data from its dependencies
    """
    save_when = SaveWhen.EXPLICIT

    def infer_dtype(self):
        deps_by_kind = self.dependencies_by_kind()
        if len(deps_by_kind) != 1:
            raise ValueError("MergeOnlyPlugins can only merge data "
                             "of the same kind, but got multiple kinds: "
                             + str(deps_by_kind))

        return strax.merged_dtype([self.deps[d].dtype
                                   for d in self.depends_on])

    def compute(self, **kwargs):
        return kwargs[list(kwargs.keys())[0]]



import glob
import os
import time
import shutil

import numpy as np
import numba

import strax
from .common import to_pe
export, __all__ = strax.exporter()


@export
@strax.takes_config(
    strax.Option('input_dir', type=str, track=False,
                 help="Directory where readers put data"),
    strax.Option('erase', default=False, track=False,
                 help="Delete reader data after processing"))
class DAQReader(strax.Plugin):
    provides = 'raw_records'
    depends_on = tuple()
    dtype = strax.record_dtype()

    parallel = 'process'
    rechunk_on_save = False

    def _path(self, chunk_i):
        return self.config["input_dir"] + f'/{chunk_i:06d}'

    def _chunk_paths(self, chunk_i):
        """Return paths to previous, current and next chunk
        If any of them does not exist, their path is replaced by False.
        """
        p = self._path(chunk_i)
        return tuple([
            q if os.path.exists(q) else False
            for q in [p + '_pre', p, p + '_post']])

    def check_next_ready_or_done(self, chunk_i):
        while True:
            ended = os.path.exists(self.config["input_dir"] + f'/THE_END')
            pre, current, post = self._chunk_paths(chunk_i)
            next_ahead = os.path.exists(self._path(chunk_i + 1))
            if (current and (
                    (pre and post
                     or chunk_i == 0 and post
                     or ended and (pre and not next_ahead)))):
                return True
            if ended and not current:
                return False
            print(f"Waiting for chunk {chunk_i}, sleeping")
            time.sleep(2)

    def _load_chunk(self, path, kind='central'):
        records = [strax.load_file(fn,
                                   compressor='blosc',
                                   dtype=strax.record_dtype())
                   for fn in glob.glob(f'{path}/reader_*')]
        records = np.concatenate(records)
        records = strax.sort_by_time(records)
        if kind == 'central':
            return records
        result = strax.from_break(records,
                                  safe_break=int(1e3),  # TODO config?
                                  left=kind == 'post',
                                  tolerant=True)
        if self.config['erase']:
            shutil.rmtree(path)
        return result

    def compute(self, chunk_i):
        pre, current, post = self._chunk_paths(chunk_i)
        records = np.concatenate(
            ([self._load_chunk(pre, kind='pre')] if pre else [])
            + [self._load_chunk(current)]
            + ([self._load_chunk(post, kind='post')] if post else [])
        )

        strax.baseline(records)
        strax.integrate(records)

        timespan_sec = (records[-1]['time'] - records[0]['time']) / 1e9
        print(f'{chunk_i}: read {records.nbytes/1e6:.2f} MB '
              f'({len(records)} records, '
              f'{timespan_sec:.2f} sec) from readers')

        return records


@export
class Records(strax.Plugin):
    depends_on = ('raw_records',)
    data_kind = 'records'   # TODO: indicate cuts have been done?
    compressor = 'zstd'
    parallel = True
    rechunk_on_save = False
    dtype = strax.record_dtype()

    def compute(self, raw_records):
        r = strax.exclude_tails(raw_records, to_pe)
        hits = strax.find_hits(r)
        strax.cut_outside_hits(r, hits)
        return r


@export
@strax.takes_config(
    strax.Option('diagnose_sorting', track=False, default=False,
                 help="Enable runtime checks for sorting and disjointness"))
class Peaks(strax.Plugin):
    depends_on = ('records',)
    data_kind = 'peaks'
    parallel = True
    rechunk_on_save = False
    dtype = strax.peak_dtype(n_channels=len(to_pe))

    def compute(self, records):
        r = records
        hits = strax.find_hits(r)       # TODO: Duplicate work
        hits = strax.sort_by_time(hits)

        peaks = strax.find_peaks(hits, to_pe,
                                 result_dtype=self.dtype)
        strax.sum_waveform(peaks, r, to_pe)

        peaks = strax.split_peaks(peaks, r, to_pe)

        strax.compute_widths(peaks)

        if self.config['diagnose_sorting']:
            assert np.diff(r['time']).min() >= 0, "Records not sorted"
            assert np.diff(hits['time']).min() >= 0, "Hits not sorted"
            assert np.all(peaks['time'][1:]
                          >= strax.endtime(peaks)[:-1]), "Peaks not disjoint"

        return peaks


@export
class PeakBasics(strax.Plugin):
    parallel = True
    depends_on = ('peaks',)
    dtype = [
        (('Start time of the peak (ns since unix epoch)',
          'time'), np.int64),
        (('End time of the peak (ns since unix epoch)',
          'endtime'), np.int64),
        (('Peak integral in PE',
            'area'), np.float32),
        (('Number of PMTs contributing to the peak',
            'n_channels'), np.int16),
        (('PMT number which contributes the most PE',
            'max_pmt'), np.int16),
        (('Width (in ns) of the central 50% area of the peak',
            'range_50p_area'), np.float32),
        (('Fraction of area seen by the top array',
            'area_fraction_top'), np.float32),

        (('Length of the peak waveform in samples',
          'length'), np.int32),
        (('Time resolution of the peak waveform in ns',
          'dt'), np.int16),
    ]

    def compute(self, peaks):
        p = peaks
        r = np.zeros(len(p), self.dtype)
        for q in 'time length dt area'.split():
            r[q] = p[q]
        r['endtime'] = p['time'] + p['dt'] * p['length']
        r['n_channels'] = (p['area_per_channel'] > 0).sum(axis=1)
        r['range_50p_area'] = p['width'][:, 5]
        r['max_pmt'] = np.argmax(p['area_per_channel'], axis=1)

        # TODO: get n_top_pmts from config...
        area_top = (p['area_per_channel'][:, :127]
                    * to_pe[:127].reshape(1, -1)).sum(axis=1)
        # Negative-area peaks get 0 AFT - TODO why not NaN?
        m = p['area'] > 0
        r['area_fraction_top'][m] = area_top[m]/p['area'][m]
        return r


@export
@strax.takes_config(
    strax.Option('s1_max_width', default=150,
                 help="Maximum (IQR) width of S1s"),
    strax.Option('s1_min_n_channels', default=3,
                 help="Minimum number of PMTs that must contribute to a S1"),
    strax.Option('s2_min_area', default=100,
                 help="Minimum area (PE) for S2s"),
    strax.Option('s2_min_width', default=200,
                 help="Minimum width for S2s"))
class PeakClassification(strax.Plugin):
    __version__ = '0.0.1'
    depends_on = ('peak_basics',)
    dtype = [
        ('type', np.int8, 'Classification of the peak.')]
    parallel = True

    def compute(self, peaks):
        p = peaks
        r = np.zeros(len(p), dtype=self.dtype)

        is_s1 = p['n_channels'] > self.config['s1_min_n_channels']
        is_s1 &= p['range_50p_area'] < self.config['s1_max_width']
        r['type'][is_s1] = 1

        is_s2 = p['area'] > self.config['s2_min_area']
        is_s2 &= p['range_50p_area'] > self.config['s2_min_width']
        r['type'][is_s2] = 2

        return r


@strax.takes_config(
    strax.Option('min_area_fraction', default=0.5,
                 help='The area of competing peaks must be at least '
                      'this fraction of that of the considered peak'),
    strax.Option('nearby_window', default=int(1e7),
                 help='Peaks starting within this time window (on either side)'
                      'in ns count as nearby.'),
    strax.Option('ignore_below', default=10,
                 help='Peaks smaller than this are ignored completely. '
                      'This is necessary to have "safe breaks" in the data.')
)
class NCompeting(strax.Plugin):
    depends_on = ('peak_basics',)
    dtype = [
        ('n_competing', np.int32,
            'Number of nearby similarly-sized (or larger) peaks')]

    def rechunk_input(self, iters):
        return dict(peaks=strax.chunk_by_break(
            iters['peaks'],
            safe_break=self.config['nearby_window'],
            ignore_below=self.config['ignore_below']
        ))

    def compute(self, peaks):
        # TODO: allow dict of arrays output
        result = np.zeros(len(peaks), dtype=self.dtype)
        result['n_competing'] = find_n_competing(
            peaks,
            window=self.config['nearby_window'],
            fraction=self.config['min_area_fraction'],
            ignore_below=self.config['ignore_below'])
        return result


@numba.jit(nopython=True, nogil=True, cache=True)
def find_n_competing(peaks, window, fraction, ignore_below):
    n = len(peaks)
    t = peaks['time']
    a = peaks['area']
    results = np.zeros(n, dtype=np.int32)

    left_i = 0
    right_i = 0
    for i, peak in enumerate(peaks):
        while t[left_i] + window < t[i] and left_i < n - 1:
            left_i += 1
        while t[right_i] - window < t[i] and right_i < n - 1:
            right_i += 1
        results[i] = np.sum(a[left_i:right_i + 1] > max(ignore_below,
                                                        a[i] * fraction))

    return results - 1


def find_peak_groups(peaks, gap_threshold,
                     left_extension=0, right_extension=0,
                     max_duration=int(1e9)):
    # Mock up a "hits" array so we can just use the existing peakfinder
    # It doesn't work on raw peaks, since they might have different dts
    # TODO: is there no cleaner way?
    fake_hits = np.zeros(len(peaks), dtype=strax.hit_dtype)
    fake_hits['dt'] = 1
    fake_hits['time'] = peaks['time']
    # TODO: could this cause int nonsense?
    fake_hits['length'] = peaks['endtime'] - peaks['time']
    fake_peaks = strax.find_peaks(
        fake_hits, to_pe=np.zeros(1),
        gap_threshold=gap_threshold,
        left_extension=left_extension, right_extension=right_extension,
        min_hits=1, min_area=0,
        max_duration=max_duration)
    # TODO: cleanup input of meaningless fields?
    # (e.g. sum waveform)
    return fake_peaks


@strax.takes_config(
    strax.Option('trigger_threshold', default=100,
                 help='Peaks with area (PE) below this do NOT cause events'),
    strax.Option('max_competing', default=7,
                 help='Peaks with  this number of similarly-sized or higher '
                      'peaks do NOT cause events'),
    strax.Option('left_extension', default=int(1e6),
                 help='Extend events this many ns to the left from each '
                      'triggering peak'),
    strax.Option('right_extension', default=int(1e6),
                 help='Extend events this many ns to the right from each '
                      'triggering peak'),
    strax.Option('max_event_duration', default=int(1e7),
                 help='Events longer than this are forcefully ended, '
                      'triggers in the truncated part are lost!'),
)
class Events(strax.Plugin):
    depends_on = ['peak_basics', 'n_competing']
    data_kind = 'events'
    dtype = [
        ('event_number', np.int64, 'Event number in this dataset'),
        ('time', np.int64, 'Event start time in ns since the unix epoch'),
        ('endtime', np.int64, 'Event end time in ns since the unix epoch')]
    parallel = False    # Since keeping state (events_seen)
    events_seen = 0

    def rechunk_input(self, iters):
        return dict(peaks=strax.chunk_by_break(
            iters['peaks'],
            safe_break=(self.config['left_extension']
                        + self.config['right_extension'] + 1),
            ignore_below=self.config['trigger_threshold']
        ))

    def compute(self, peaks):
        le = self.config['left_extension']
        re = self.config['right_extension']

        triggers = peaks[
            (peaks['area'] > self.config['trigger_threshold'])
            & (peaks['n_competing'] <= self.config['max_competing'])
            ]

        # Join nearby triggers
        peak_groups = find_peak_groups(
            triggers,
            gap_threshold=le + re + 1,
            left_extension=le,
            right_extension=re,
            max_duration=self.config['max_event_duration'])

        result = np.zeros(len(peak_groups), self.dtype)
        result['time'] = peak_groups['time']
        result['endtime'] = (peak_groups['time']
                             + peak_groups['length'] * peak_groups['dt'])
        result['event_number'] = (np.arange(len(result))
                                  + self.events_seen)

        # Filter out events without S1 + S2
        self.events_seen += len(result)

        return result
        # TODO: someday investigate why loopplugin doesn't give
        # anything if events do not contain peaks..


@export
class EventBasics(strax.LoopPlugin):
    depends_on = ('events',
                  'peak_basics', 'peak_classification', 'n_competing')

    def infer_dtype(self):
        dtype = [(('Number of peaks in the event',
                   'n_peaks'), np.int32),
                 (('Drift time between main S1 and S2 in ns',
                   'drift_time'), np.int64)]
        for i in [1, 2]:
            dtype += [((f'Main S{i} peak index',
                        f's{i}_index'), np.int32),
                      ((f'Main S{i} area (PE)',
                        f's{i}_area'), np.float32),
                      ((f'Main S{i} area fraction top',
                        f's{i}_area_fraction_top'), np.float32),
                      ((f'Main S{i} width (ns, 50% area)',
                        f's{i}_range_50p_area'), np.float32),
                      ((f'Main S{i} number of competing peaks',
                        f's{i}_n_competing'), np.int32)]
        return dtype

    def compute_loop(self, event, peaks):
        result = dict(n_peaks=len(peaks))
        if not len(peaks):
            return result

        main_s = dict()
        for s_i in [1, 2]:
            s_mask = peaks['type'] == s_i
            ss = peaks[s_mask]
            s_indices = np.arange(len(peaks))[s_mask]
            if not len(ss):
                result[f's{s_i}_index'] = -1
                continue
            main_i = np.argmax(ss['area'])
            result[f's{s_i}_index'] = s_indices[main_i]
            s = main_s[s_i] = ss[main_i]
            for prop in ['area', 'area_fraction_top',
                         'range_50p_area', 'n_competing']:
                result[f's{s_i}_{prop}'] = s[prop]

        if len(main_s) == 2:
            result['drift_time'] = main_s[2]['time'] - main_s[1]['time']

        return result


# LargestS2Area is just an example plugin
# The same info is provided in event_basics

class LargestS2Area(strax.LoopPlugin):
    """Find the largest S2 area in the event.
    """
    __version__ = '0.1'
    depends_on = ('events', 'peak_basics', 'peak_classification')

    dtype = [
        ('largest_s2_area', np.float32,
            'Area (PE) of largest S2 in event')]

    def compute_loop(self, event, peaks):

        s2s = peaks[peaks['type'] == 2]

        result = 0
        if len(s2s):
            result = s2s['area'].max()

        return dict(largest_s2_area=result)



"""Convert pax .zip files to flat records format
"""
import numpy as np
import os
import glob

import strax
export, __all__ = strax.exporter()


@export
def pax_to_records(input_filename,
                   samples_per_record=strax.DEFAULT_RECORD_LENGTH):
    """Return pulse records array from pax zip input_filename"""
    from pax import core   # Pax is not a dependency
    mypax = core.Processor('XENON1T', config_dict=dict(
            pax=dict(
                look_for_config_in_runs_db=False,
                plugin_group_names=['input'],
                encoder_plugin=None,
                input_name=input_filename),
            # Fast startup: skip loading big maps
            WaveformSimulator=dict(
                s1_light_yield_map='placeholder_map.json',
                s2_light_yield_map='placeholder_map.json',
                s1_patterns_file=None,
                s2_patterns_file=None)))

    def get_events():
        for e in mypax.get_events():
            yield mypax.process_event(e)

    # We loop over the events twice for convenience
    # Yeah, this is probably not optimal
    pulse_lengths = np.array([p.length
                              for e in get_events()
                              for p in e.pulses])

    n_records = strax.records_needed(pulse_lengths, samples_per_record).sum()
    records = np.zeros(n_records,
                       dtype=strax.record_dtype(samples_per_record))

    output_record_index = 0  # Record offset in data
    for event in get_events():
        for p in event.pulses:

            n_records = strax.records_needed(p.length, samples_per_record)
            for rec_i in range(n_records):
                r = records[output_record_index]
                r['time'] = (event.start_time
                             + p.left * 10
                             + rec_i * samples_per_record * 10)
                r['channel'] = p.channel
                r['pulse_length'] = p.length
                r['record_i'] = rec_i
                r['dt'] = 10

                # How much are we storing in this record?
                if rec_i != n_records - 1:
                    # There's more chunks coming, so we store a full chunk
                    n_store = samples_per_record
                    assert p.length > samples_per_record * (rec_i + 1)
                else:
                    # Just enough to store the rest of the data
                    # Note it's not p.length % samples_per_record!!!
                    # (that would be zero if we have to store a full record)
                    n_store = p.length - samples_per_record * rec_i

                assert 0 <= n_store <= samples_per_record
                r['length'] = n_store

                offset = rec_i * samples_per_record
                r['data'][:n_store] = p.raw_data[offset:offset + n_store]
                output_record_index += 1

    mypax.shutdown()

    # In strax data, records are always stored
    # sorted, baselined and integrated
    records = strax.sort_by_time(records)
    strax.baseline(records)
    strax.integrate(records)
    return records


@export
@strax.takes_config(
    strax.Option('pax_raw_dir', default='/data/xenon/raw', track=False,
                 help="Directory with raw pax datasets"),
    strax.Option('stop_after_zips', default=0, track=False,
                 help="Convert only this many zip files. 0 = all.")
)
class RecordsFromPax(strax.Plugin):
    provides = 'raw_records'
    data_kind = 'raw_records'
    depends_on = tuple()
    dtype = strax.record_dtype()
    parallel = False

    def iter(self, *args, **kwargs):
        if not os.path.exists(self.config['pax_raw_dir']):
            raise FileNotFoundError(self.config['pax_raw_dir'])
        input_dir = os.path.join(self.config['pax_raw_dir'], self.run_id)
        pax_files = sorted(glob.glob(input_dir + '/*.zip'))
        pax_sizes = np.array([os.path.getsize(x)
                              for x in pax_files])
        print(f"Found {len(pax_files)} files, {pax_sizes.sum() / 1e9:.2f} GB")
        for file_i, in_fn in enumerate(pax_files):
            if (self.config['stop_after_zips']
                    and file_i >= self.config['stop_after_zips']):
                break
            yield strax.xenon.pax_interface.pax_to_records(in_fn)



import builtins
import collections
import logging
import inspect
import fnmatch
import typing as ty
import warnings
import random
import string

import numpy as np
import pandas as pd

import strax
export, __all__ = strax.exporter()

get_docs = """
    :param run_id: run id to get
    :param target: data type to get
    :param save: data types you would like to save
    to cache, if they occur in intermediate computations.
    Some plugins save automatically.
    :param max_workers: Number of worker threads/processes to spawn.
    In practice more CPUs may be used due to strax's multithreading.
"""


@export
class Context:
    """Context for strax analysis.

    A context holds info on HOW to process data, such as which plugins provide
    what data types, where to store which results, and configuration options
    for the plugins.

    You start all strax processing through a context.
    """

    def __init__(self,
                 storage=None,
                 config=None,
                 register=None,
                 register_all=None):
        self.log = logging.getLogger('strax')

        if storage is None:
            storage = ['./strax_data']
        if not isinstance(storage, (list, tuple)):
            storage = [storage]
        self.storage = [strax.FileStore(s) if isinstance(s, str) else s
                        for s in storage]

        self._plugin_class_registry = dict()
        self._plugin_instance_cache = dict()

        self.set_config(config, mode='new')

        if register is not None:
            self.register(register)
        if register_all is not None:
            self.register_all(register_all)

    def new_context(self,
                    storage=tuple(),
                    config=None,
                    register=None,
                    register_all=None,
                    replace=False):
        """Return a new context with new setting adding to those in
        this context.
        :param replace: If True, replaces settings rather than adding them.
        """
        if not isinstance(storage, (list, tuple)):
            storage = [storage]
        if config is None:
            config = dict()
        if register is None:
            register = []
        if not isinstance(register, (tuple, list)):
            register = [register]

        if not replace:
            storage = self.storage + list(storage)
            config = {**self.config, **config}
            register = list(self._plugin_class_registry.values()) + register

        return Context(storage=storage,
                       config=config,
                       register=register,
                       register_all=register_all)

    def set_config(self, config=None, mode='update'):
        """Set new configuration options

        :param config: dict of new options
        :param mode: can be either
         - update: Add to or override current options in context
         - setdefault: Add to current options, but do not override
         - replace: Erase config form this context, then set only these options
        """
        if config is None:
            config = dict()
        if mode == 'update':
            self.config.update(config)
        elif mode == 'setdefault':
            for k in config:
                self.config.setdefault(k, config[k])
        elif mode == 'new':
            self.config = config
        else:
            raise RuntimeError("Expected update, setdefault or new as config"
                               " setting mode")

    def register(self, plugin_class, provides=None):
        """Register plugin_class as provider for data types in provides.
        :param plugin_class: class inheriting from StraxPlugin
        :param provides: list of data types which this plugin provides.

        Plugins always register for the data type specified in the .provide
        class attribute. If such is not available, we will construct one from
        the class name (CamelCase -> snake_case)

        Returns plugin_class (so this can be used as a decorator)
        """
        if isinstance(plugin_class, (tuple, list)) and provides is None:
            # shortcut for multiple registration
            # TODO: document
            for x in plugin_class:
                self.register(x)
            return

        if not hasattr(plugin_class, 'provides'):
            # No output name specified: construct one from the class name
            snake_name = strax.camel_to_snake(plugin_class.__name__)
            plugin_class.provides = snake_name

        if provides is not None:
            provides += [plugin_class.provides]
        else:
            provides = [plugin_class.provides]

        for p in provides:
            self._plugin_class_registry[p] = plugin_class

        return plugin_class

    def search_field(self, pattern):
        cache = dict()
        for d in self._plugin_class_registry:
            if d not in cache:
                cache.update(self._get_plugins((d,)))
            p = cache[d]

            for field_name in p.dtype.names:
                if fnmatch.fnmatch(field_name, pattern):
                    print(f"{field_name} is part of {d} "
                          f"(provided by {p.__class__.__name__})")

    def search_config(self, data_type, pattern='*'):
        r = []
        for d, p in self._get_plugins((data_type,)).items():
            for opt in p.takes_config.values():
                if not fnmatch.fnmatch(opt.name, pattern):
                    continue
                try:
                    default = opt.get_default()
                except strax.InvalidConfiguration:
                    default = '<OMITTED>'
                r.append(dict(
                    option=opt.name,
                    default=default,
                    current=self.config.get(opt.name, '<OMITTED>'),
                    data_type=d,
                    help=opt.help))
        if len(r):
            return pd.DataFrame(r, columns=r[0].keys())
        return pd.DataFrame([])

    def lineage(self, data_type):
        return self._get_plugins((data_type,))[data_type].lineage

    def register_all(self, module):
        """Register all plugins defined in module"""
        if isinstance(module, (tuple, list)):
            # Secret shortcut for multiple registration
            for x in module:
                self.register_all(x)
            return

        for x in dir(module):
            x = getattr(module, x)
            if type(x) != type(type):
                continue
            if issubclass(x, strax.Plugin):
                self.register(x)

    def data_info(self, data_name: str) -> pd.DataFrame:
        """Return pandas DataFrame describing fields in data_name"""
        p = self._get_plugins((data_name,))[data_name]
        display_headers = ['Field name', 'Data type', 'Comment']
        result = []
        for name, dtype in strax.utils.unpack_dtype(p.dtype):
            if isinstance(name, tuple):
                title, name = name
            else:
                title = ''
            result.append([name, dtype, title])
        return pd.DataFrame(result, columns=display_headers)

    def _set_plugin_config(self, p, tolerant=True):
        # Explicit type check, since if someone calls this with
        # plugin CLASSES, funny business might ensue
        assert isinstance(p, strax.Plugin)
        config = self.config.copy()
        for opt in p.takes_config.values():
            try:
                opt.validate(config)
            except strax.InvalidConfiguration:
                if not tolerant:
                    raise
        p.config = {k: v for k, v in config.items()
                    if k in p.takes_config}

    def _get_plugins(self,
                     targets: ty.Tuple[str] = tuple(),
                     run_id: str = 'UNKNOWN'
                     ) -> ty.Dict[str, strax.Plugin]:
        """Return dictionary of plugins necessary to compute targets
        from scratch.
        """
        # Check all config options are taken by some registered plugin class
        # (helps spot typos)
        all_opts = set().union(*[
            pc.takes_config.keys()
            for pc in self._plugin_class_registry.values()])
        for k in self.config:
            if k not in all_opts:
                warnings.warn(f"Option {k} not taken by any registered plugin")

        # Initialize plugins for the entire computation graph
        # (most likely far further down than we need)
        # to get lineages and dependency info.
        def get_plugin(d):
            nonlocal plugins

            if d not in self._plugin_class_registry:
                raise KeyError(f"No plugin class registered that provides {d}")

            p = self._plugin_class_registry[d]()
            p.run_id = run_id

            # The plugin may not get all the required options here
            # but we don't know if we need the plugin yet
            self._set_plugin_config(p, tolerant=True)

            # TODO: check can now be moved inside plugin
            compute_pars = list(
                inspect.signature(p.compute).parameters.keys())
            if 'chunk_i' in compute_pars:
                p.compute_takes_chunk_i = True
                del compute_pars[compute_pars.index('chunk_i')]

            plugins[d] = p

            p.deps = {d: get_plugin(d) for d in p.depends_on}

            p.lineage = {d: (p.__class__.__name__,
                             p.version(run_id),
                             {k: v for k, v in p.config.items()
                              if p.takes_config[k].track})}
            for d in p.depends_on:
                p.lineage.update(p.deps[d].lineage)

            if not hasattr(p, 'data_kind'):
                if len(p.depends_on):
                    # Assume data kind is the same as the first dependency
                    p.data_kind = p.deps[p.depends_on[0]].data_kind
                else:
                    # No dependencies: assume provided data kind and
                    # data type are synonymous
                    p.data_kind = p.provides

            if not hasattr(p, 'dtype'):
                p.dtype = p.infer_dtype()

            if not isinstance(p, np.dtype):
                dtype = []
                for x in p.dtype:
                    if len(x) == 3:
                        if isinstance(x[0], tuple):
                            # Numpy syntax for array field
                            dtype.append(x)
                        else:
                            # Lazy syntax for normal field
                            field_name, field_type, comment = x
                            dtype.append(((comment, field_name), field_type))
                    elif len(x) == 2:
                        # (field_name, type)
                        dtype.append(x)
                    elif len(x) == 1:
                        # Omitted type: assume float
                        dtype.append((x, np.float))
                    else:
                        raise ValueError(f"Invalid field specification {x}")
                p.dtype = np.dtype(dtype)
            return p

        plugins = collections.defaultdict(get_plugin)
        for t in targets:
            # This works without the RHS too, but your IDE might not get it :-)
            plugins[t] = get_plugin(t)

        return plugins

    def get_components(self, run_id: str,
                       targets=tuple(), save=tuple()
                       ) -> strax.ProcessorComponents:
        """Return components for setting up a processor
        {get_docs}
        """
        save = strax.to_str_tuple(save)
        targets = strax.to_str_tuple(targets)

        plugins = self._get_plugins(targets, run_id)

        # Get savers/loaders, and meanwhile filter out plugins that do not
        # have to do computation.(their instances will stick around
        # though the .deps attribute of plugins that do)
        loaders = dict()
        savers = collections.defaultdict(list)
        seen = set()
        to_compute = dict()

        def check_cache(d):
            nonlocal plugins, loaders, savers, seen
            if d in seen:
                return
            seen.add(d)
            p = plugins[d]
            key = strax.CacheKey(run_id, d, p.lineage)

            for sb_i, sb in enumerate(self.storage):
                try:
                    loaders[d] = sb.loader(key)
                    # Found it! No need to make it or save it
                    del plugins[d]
                    return
                except strax.NotCached:
                    continue

            # Not in any cache. We will be computing it.
            to_compute[d] = p
            for dep_d in p.depends_on:
                check_cache(dep_d)

            # We're making this OR it gets fed in. Should we save it?
            if p.save_when == strax.SaveWhen.NEVER:
                if d in save:
                    raise ValueError("Plugin forbids saving of {d}")
                return
            elif p.save_when == strax.SaveWhen.TARGET:
                if d != targets:
                    return
            elif p.save_when == strax.SaveWhen.EXPLICIT:
                if d not in save:
                    return
            else:
                assert p.save_when == strax.SaveWhen.ALWAYS

            for sb_i, sb in enumerate(self.storage):
                if not sb.provides(d, write=True):
                    continue
                s = sb.saver(key, p.metadata(run_id))
                s.meta_only = p.save_meta_only
                savers[d].append(s)

        for d in targets:
            check_cache(d)
        plugins = to_compute

        intersec = list(plugins.keys() & loaders.keys())
        if len(intersec):
            raise RuntimeError("{intersec} both computed and loaded?!")

        # Check all required options are available / set defaults
        for p in plugins.values():
            self._set_plugin_config(p, tolerant=False)
        return strax.ProcessorComponents(
            plugins=plugins,
            loaders=loaders,
            savers=dict(savers),
            targets=targets)

    def get_iter(self, run_id: str, targets, save=tuple(), max_workers=None,
                 **kwargs) -> ty.Iterator[np.ndarray]:
        """Compute target for run_id and iterate over results
        {get_docs}
        TODO: This is not quite a normal iterator: if you break
        results will still accumulate in a background thread!
        """
        if len(kwargs):
            self = self.new_context(**kwargs)
        # If multiple targets of the same kind, create a MergeOnlyPlugin
        # automatically
        if isinstance(targets, (list, tuple)) and len(targets) > 1:
            plugins = self._get_plugins(targets=targets)
            if len(set(plugins[d].data_kind for d in targets)) == 1:
                temp_name = ''.join(random.choices(
                    string.ascii_lowercase, k=10))
                temp_merge = type(temp_name,
                                  (strax.MergeOnlyPlugin,),
                                  dict(depends_on=tuple(targets)))
                self.register(temp_merge)
                targets = temp_name
                # TODO: auto-unregister? Better to have a temp register
                # override option in get_components
            else:
                raise RuntimeError("Cannot automerge different data kinds!")

        components = self.get_components(run_id, targets=targets, save=save)
        yield from strax.ThreadedMailboxProcessor(
            components, max_workers=max_workers).iter()

    def make(self, *args, **kwargs) -> None:
        """Compute target for run_id (return nothing)
        {get_docs}
        """
        for _ in self.get_iter(*args, **kwargs):
            pass

    def get_array(self, *args, **kwargs) -> np.ndarray:
        """Compute target for run_id and return as numpy array
        {get_docs}
        """
        return np.concatenate(list(self.get_iter(*args, **kwargs)))

    def get_df(self, *args, **kwargs) -> pd.DataFrame:
        """Compute target for run_id and return as pandas DataFrame
        {get_docs}
        """
        return pd.DataFrame.from_records(self.get_array(*args, **kwargs))

    def get_meta(self, run_id, target) -> dict:
        """Return metadata for target for run_id, or raise NotCached
        if data is not yet available.
        {get_docs}
        """
        p = self._get_plugins((target,))[target]
        key = strax.CacheKey(run_id, target, p.lineage)
        for sb in self.storage:
            if sb.has(key):
                return sb.load_meta(key)
        raise strax.NotCached(f"Can't load metadata, "
                              f"data for {key} not available")


##
# Config specification. Maybe should be its own file?
##


@export
class InvalidConfiguration(Exception):
    pass


# Todo:  Does it really have to be a decorator just for nice error message?
@export
def takes_config(*options):
    def wrapped(plugin_class):
        result = []
        for opt in options:
            if isinstance(opt, str):
                opt = Option(opt)
            elif not isinstance(opt, Option):
                raise RuntimeError("Specify config options by str or Option")
            opt.taken_by = plugin_class.__name__
            result.append(opt)

        plugin_class.takes_config = {opt.name: opt for opt in result}
        return plugin_class

    return wrapped


# Use instead of None since None might be a proper value/default
OMITTED = 'Argument was not given'


@export
class Option:
    taken_by = "UNKNOWN???"

    def __init__(self,
                 name: str,
                 type: type = OMITTED,
                 default: ty.Any = OMITTED,
                 default_factory: ty.Callable = OMITTED,
                 track=True,
                 help: str = ''):
        self.name = name
        self.type = type
        type = builtins.type
        self.default = default
        self.track = track
        self.default_factory = default_factory
        self.help = help

        if (self.default is not OMITTED
                and self.default_factory is not OMITTED):
            raise RuntimeError(f"Tried to specify both default and "
                               f"default_factory for option {self.name}")

        if type is OMITTED and default is not OMITTED:
            self.type = type(default)

    def get_default(self):
        if self.default is OMITTED:
            if self.default_factory is OMITTED:
                raise InvalidConfiguration(f"Missing option {self.name} "
                                           f"required by {self.taken_by}")
            return self.default_factory()
        return self.default

    def validate(self, config, set_defaults=True):
        """Checks if the option is in config and sets defaults if needed.
        """
        if self.name in config:
            value = config[self.name]
            if (self.type is not OMITTED
                    and not isinstance(value, self.type)):
                raise InvalidConfiguration(
                    f"Invalid type for option {self.name}. "
                    f"Excepted a {self.type}, got a {type(value)}")
        elif set_defaults:
            config[self.name] = self.get_default()



import contextlib
from functools import wraps
import re
import typing
from hashlib import sha1
import pickle

import numpy as np
import numba
import dill


# Change numba's caching backend from pickle to dill
# I'm sure they don't mind...
# Otherwise we get strange errors while caching the @growing_result functions
numba.caching.pickle = dill


def exporter(export_self=False):
    """Export utility modified from https://stackoverflow.com/a/41895194
    Returns export decorator, __all__ list
    """
    all_ = []
    if export_self:
        all_.append('exporter')

    def decorator(obj):
        all_.append(obj.__name__)
        return obj

    return decorator, all_


export, __all__ = exporter(export_self=True)


@export
def records_needed(pulse_length, samples_per_record):
    """Return records needed to store pulse_length samples"""
    return 1 + (pulse_length - 1) // samples_per_record


@export
def growing_result(dtype=np.int, chunk_size=10000):
    """Decorator factory for functions that fill numpy arrays
    Functions must obey following API:
     - accept _result_buffer keyword argument with default None;
       this will be the buffer array of specified dtype and length chunk_size
       (it's an optional argument so this decorator preserves signature)
     - 'yield N' from function will cause first elements to be saved
     - function is responsible for tracking offset, calling yield on time,
       and clearing the buffer afterwards.
     - optionally, accept result_dtype argument with default None;
       this allows function user to specify return dtype

    See test_utils.py for a simple example (I can't get it to run as a doctest
    unfortunately)
    """
    def _growing_result(f):
        @wraps(f)
        def wrapped_f(*args, **kwargs):

            if '_result_buffer' in kwargs:
                raise ValueError("_result_buffer argument is internal-only")

            _dtype = kwargs.get('result_dtype')
            if _dtype is None:
                _dtype = dtype
            else:
                del kwargs['result_dtype']   # Don't pass it on to function
            buffer = np.zeros(chunk_size, _dtype)

            # Keep a list of saved buffers to concatenate at the end
            saved_buffers = []
            for n_added in f(*args, _result_buffer=buffer, **kwargs):
                saved_buffers.append(buffer[:n_added].copy())

            # If nothing returned, return an empty array of the right dtype
            if not len(saved_buffers):
                return np.zeros(0, _dtype)
            return np.concatenate(saved_buffers)

        return wrapped_f

    return _growing_result


@export
def unpack_dtype(dtype):
    """Return list of tuples needed to construct the dtype

    dtype == np.dtype(unpack_dtype(dtype))
    """
    result = []
    fields = dtype.fields
    for field_name in dtype.names:
        fieldinfo = fields[field_name]
        if len(fieldinfo) == 3:
            # The field has a "titles" attribute.
            # In this case, the tuple returned by .fields is inconsistent
            # with the tuple expected by np.dtype constructor :-(
            field_dtype, some_number, field_title = fieldinfo
            result.append(((field_title, field_name), field_dtype))
        else:
            field_dtype, some_number = fieldinfo
            result.append((field_name, field_dtype))
    return result


@export
def merged_dtype(dtypes):
    result = {}
    for x in dtypes:
        for unpacked_dtype in unpack_dtype(x):
            field_name = unpacked_dtype[0]
            if isinstance(field_name, tuple):
                field_name = field_name[0]

            if field_name in result:
                # Name collision
                continue

            result[field_name] = unpacked_dtype

    return list(result.values())


@export
def merge_arrs(arrs):
    """Merge structured arrays of equal length.

    On field name collisions, data from later arrays is kept.
    """
    # Much faster than the similar function in numpy.lib.recfunctions

    n = len(arrs[0])
    if not all([len(x) == n for x in arrs]):
        raise ValueError("Arrays must all have the same length")

    result = np.zeros(n, dtype=merged_dtype([x.dtype for x in arrs]))
    for arr in arrs:
        for fn in arr.dtype.names:
            result[fn] = arr[fn]
    return result


@export
def camel_to_snake(x):
    """Convert x from CamelCase to snake_case"""
    # From https://stackoverflow.com/questions/1175208
    x = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', x)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', x).lower()


@export
@contextlib.contextmanager
def profile_threaded(filename):
    import yappi            # noqa   # yappi is not a dependency
    if filename is None:
        yield
        return

    yappi.start()
    yield
    yappi.stop()
    p = yappi.get_func_stats()
    p = yappi.convert2pstats(p)
    p.dump_stats(filename)
    yappi.clear_stats()


@export
def to_str_tuple(x) -> typing.Tuple[str]:
    if isinstance(x, str):
        return x,
    elif isinstance(x, list):
        return tuple(x)
    return x


@export
def hashablize(obj):
    """Convert a container hierarchy into one that can be hashed.
    See http://stackoverflow.com/questions/985294
    """
    try:
        hash(obj)
    except TypeError:
        if isinstance(obj, dict):
            return tuple((k, hashablize(v)) for (k, v) in sorted(obj.items()))
        elif isinstance(obj, np.ndarray):
            return tuple(obj.tolist())
        elif hasattr(obj, '__iter__'):
            return tuple(hashablize(o) for o in obj)
        else:
            raise TypeError("Can't hashablize object of type %r" % type(obj))
    else:
        return obj


@export
def deterministic_hash(thing):
    """Return a deterministic hash of a container hierarchy using hashablize,
    pickle and sha1"""
    return sha1(pickle.dumps(hashablize(thing))).hexdigest()



from concurrent import futures
import logging
import typing as ty
import itertools

import strax
export, __all__ = strax.exporter()


@export
class ProcessorComponents(ty.NamedTuple):
    """Specification to assemble a processor"""
    plugins: ty.Dict[str, strax.Plugin]
    loaders: ty.Dict[str, ty.Iterator]
    savers:  ty.Dict[str, ty.List[strax.Saver]]
    targets: ty.Tuple[str]


@export
class ThreadedMailboxProcessor:
    mailboxes: ty.Dict[str, strax.Mailbox]

    def __init__(self, components: ProcessorComponents, max_workers=None):
        self.log = logging.getLogger(self.__class__.__name__)
        self.components = components
        self.log.debug("Processor components are: " + str(components))
        plugins = components.plugins
        savers = components.savers

        # Executors for parallelization
        process_executor = futures.ProcessPoolExecutor(max_workers=max_workers)
        thread_executor = futures.ThreadPoolExecutor(max_workers=max_workers)

        # If possible, combine save and compute operations
        # so they don't have to be scheduled by executor individually.
        # This saves data transfer between cores (NUMA).
        for d, p in plugins.items():
            if not p.rechunk_on_save:
                self.log.debug(f"Putting savers for {d} in post_compute")
                for s in savers.get(d, []):
                    p.post_compute.append(s.save)
                    p.on_close.append(s.close)
                savers[d] = []

        # For the same reason, merge simple chains:
        # A -> B => A, with B as post_compute,
        # then put in plugins as B instead of A.
        # TODO: check they agree on paralellization?
        # TODO: allow compute grouping while saver does rechunk
        while True:
            for b, p_b in plugins.items():
                if (p_b.parallel and not p_b.rechunk_on_save
                        and len(p_b.depends_on) == 1
                        and b not in components.targets):
                    a = p_b.depends_on[0]
                    if a not in plugins:
                        continue
                    self.log.debug(f"Putting {b} in post_compute of {a}")
                    p_a = plugins[a]
                    p_a.post_compute.append(plugins[b].do_compute)
                    p_a.on_close.extend(p_b.on_close)
                    plugins[b] = p_a
                    del plugins[a]
                    break       # Changed plugins while iterating over it
            else:
                break

        self.log.debug("After optimization: " + str(components))

        self.mailboxes = {
            d: strax.Mailbox(name=d + '_mailbox')
            for d in itertools.chain.from_iterable(components)}

        for d, loader in components.loaders.items():
            assert d not in plugins
            self.mailboxes[d].add_sender(loader, name=f'load:{d}')

        for d, p in plugins.items():

            executor = None
            if p.parallel == 'process':
                executor = process_executor
            elif p.parallel:
                executor = thread_executor

            self.mailboxes[d].add_sender(p.iter(
                    iters={d: self.mailboxes[d].subscribe()
                           for d in p.depends_on},
                    executor=executor),
                name=f'build:{d}')

        for d, savers in savers.items():
            for s_i, saver in enumerate(savers):
                self.mailboxes[d].add_reader(saver.save_from,
                                             name=f'save_{s_i}:{d}')

    def iter(self):
        target = self.components.targets[0]
        final_generator = self.mailboxes[target].subscribe()

        self.log.debug("Starting threads")
        for m in self.mailboxes.values():
            m.start()

        self.log.debug(f"Yielding {target}")
        try:
            yield from final_generator
        except strax.MailboxKilled as e:
            self.log.debug(f"Target Mailbox ({target}) killed")
            for m in self.mailboxes.values():
                if m != target:
                    self.log.debug(f"Killing {m}")
                    m.kill(upstream=True,
                           reason=e.args[0])

        self.log.debug("Closing threads")
        for m in self.mailboxes.values():
            m.cleanup()



import numpy as np

import strax
from .common import Store, Saver, NotCached

export, __all__ = strax.exporter()

try:
    import pymongo
except ImportError:
    pymongo = None


@export
class MongoStore(Store):
    """MongoDB-based storage backend for strax, intended
    for high-level data.

    Data is stored in uncompressed / fully expanded documents (not binary
    blobs).
    A _chunk_i: integer field is added to each document.

    Metadata is stored in the same collection as data, but with a
    metadata = True attribute in the document.
    """

    def __init__(self, uri, database='strax_data', *args, **kwargs):
        """
        :param connection_uri: Connection URI, which also specifies DB
        {provides_doc}
        """
        super().__init__(*args, **kwargs)
        if pymongo is None:
            raise ImportError("Pymongo did not import: "
                              "cannot use MongoDBStore")

        self.client = pymongo.MongoClient(uri)
        self.client.admin.command('ping')
        self.db = self.client[database]

    def _find(self, key):
        if self.db[str(key)].count():
            self.log.debug(f"{key} is in cache.")
            return str(key)

        self.log.debug(f"{key} is NOT in cache.")
        raise NotCached

    def _read_chunk(self, coll_name, chunk_info, dtype, compressor):
        docs = list(self.db[coll_name].find(
            {'_chunk_i': chunk_info['chunk_i']}))

        result = np.zeros(len(docs), dtype=dtype)
        for i, d in enumerate(docs):
            for k, v in d.items():
                if k in ('_chunk_i', '_id'):
                    continue
                result[i][k] = v
        return result

    def _read_meta(self, coll_name):
        return self.db[coll_name].find_one({'metadata': True})

    def saver(self, key, metadata):
        super().saver(key, metadata)
        return MongoSaver(key, metadata, self.db[str(key)])


@export
class MongoSaver(Saver):
    prefer_rechunk = False

    def __init__(self, key, metadata, col):
        super().__init__(key, metadata)
        self.col = col
        self.col.insert(dict(metadata=True, **self.md))

    def _save_chunk(self, data, chunk_info):
        docs = [dict(_chunk_i=chunk_info['chunk_i'])
                for _ in range(len(data))]
        for i, d in enumerate(docs):
            for k in data.dtype.names:
                docs[i][k] = np.asscalar(data[i][k])

        self.col.insert(docs)
        return dict()

    def _save_chunk_metadata(self, chunk_info):
        self.col.find_one_and_update(dict(metadata=True),
                                     {'$push': {'chunks': chunk_info}})

    def close(self):
        super().close()
        update = {k: v
                  for k, v in self.md.items()
                  if k in ('writing_ended', 'exception')}
        self.col.find_one_and_update(dict(metadata=True),
                                     {'$set': update})



import os
import json
import shutil
import glob

import strax
from .common import Store, Saver, NotCached, CacheKey

export, __all__ = strax.exporter()


@export
class FileStore(Store):

    def __init__(self, data_dirs, *args, **kwargs):
        """File-based storage backend for strax.
        Data is stored in (compressed) binary dumps, metadata in json.

        :param data_dirs: List of data directories to use.
        Entries are preferred for use from first to last.

        {provides_doc}

        When writing, we save (only) in the highest-preference directory
        in which we have write permission.
        """
        super().__init__(*args, **kwargs)

        self.data_dirs = strax.to_str_tuple(data_dirs)

        for d in self.data_dirs:
            try:
                os.makedirs(d)
            except FileExistsError:
                pass
            else:
                self.log.debug(f"Created data dir {d}")

    def _find(self, key):
        for dirname in self._candidate_dirs(key):
            if os.path.exists(dirname):
                self.log.debug(f"{key} is in cache.")
                return dirname
        self.log.debug(f"{key} is NOT in cache.")
        raise NotCached

    @staticmethod
    def _key_dirname(key):
        return '_'.join([key.run_id,
                         key.data_type,
                         strax.deterministic_hash(key.lineage)])

    def _candidate_dirs(self, key: CacheKey):
        """Return directories at which data meant by key
        could be found or saved"""
        return [os.path.join(d, str(key))
                for d in self.data_dirs]

    def _read_chunk(self, dirname, chunk_info, dtype, compressor):
        fn = os.path.join(dirname, chunk_info['filename'])
        return strax.load_file(fn, dtype=dtype, compressor=compressor)

    def _read_meta(self, dirname):
        with open(dirname + '/metadata.json', mode='r') as f:
            return json.loads(f.read())

    def saver(self, key, metadata):
        super().saver(key, metadata)

        for dirname in self._candidate_dirs(key):
            # Test if the parent directory is writeable.
            # We need abspath since the dir itself may not exist,
            # even though its parent-to-be does
            parent_dir = os.path.abspath(os.path.join(dirname, os.pardir))
            if os.access(parent_dir, os.W_OK):
                self.log.debug(f"Saving {key} to {dirname}")
                break
            else:
                self.log.debug(f"{parent_dir} is not writeable, "
                               f"can't save to {dirname}")
        else:
            raise FileNotFoundError(f"No writeable directory found for {key}")

        return FileSaver(key, metadata, dirname)


@export
class FileSaver(Saver):
    """Saves data to compressed binary files

    Must work even if forked.
    Do NOT add unpickleable things as attributes (such as loggers)!
    """
    json_options = dict(sort_keys=True, indent=4)

    def __init__(self, key, metadata, dirname):
        super().__init__(key, metadata)
        self.dirname = dirname

        self.tempdirname = dirname + '_temp'
        if os.path.exists(dirname):
            print("Deleting old incomplete data in {dirname}")
            shutil.rmtree(dirname)
        if os.path.exists(self.tempdirname):
            shutil.rmtree(self.tempdirname)
        os.makedirs(self.tempdirname)

    def _save_chunk(self, data, chunk_info):
        filename = '%06d' % chunk_info['chunk_i']
        filesize = strax.save_file(
                os.path.join(self.tempdirname, filename),
                data=data,
                compressor=self.md['compressor'])
        return dict(filename=filename, filesize=filesize)

    def _save_chunk_metadata(self, chunk_info):
        if self.meta_only:
            # TODO HACK!
            chunk_info["filename"] = '%06d' % chunk_info['chunk_i']
        fn = f'{self.tempdirname}/metadata_{chunk_info["filename"]}.json'
        with open(fn, mode='w') as f:
            f.write(json.dumps(chunk_info, **self.json_options))

    def close(self):
        super().close()
        for fn in sorted(glob.glob(
                self.tempdirname + '/metadata_*.json')):
            with open(fn, mode='r') as f:
                self.md['chunks'].append(json.load(f))
            os.remove(fn)

        with open(self.tempdirname + '/metadata.json', mode='w') as f:
            f.write(json.dumps(self.md, **self.json_options))
        os.rename(self.tempdirname, self.dirname)



"""Functions that perform processing on pulses
(other than data reduction functions, which are in data_reduction.py)
"""
import numpy as np
import numba

import strax
export, __all__ = strax.exporter()

# Constant for use in record_links, to indicate there is no prev/next record
NOT_APPLICABLE = -1


@export
@numba.jit(nopython=True, nogil=True, cache=True)
def baseline(records, baseline_samples=40):
    """Subtract pulses from int(baseline), store baseline in baseline field
    :param baseline_samples: number of samples at start of pulse to average
    Assumes records are sorted in time (or at least by channel, then time)

    Assumes record_i information is accurate (so don't cut pulses before
    baselining them!)
    """
    if not len(records):
        return
    samples_per_record = len(records[0]['data'])

    # Array for looking up last baseline seen in channel
    # We only care about the channels in this set of records; a single .max()
    # is worth avoiding the hassle of passing n_channels around
    last_bl_in = np.zeros(records['channel'].max() + 1, dtype=np.int16)

    for d_i, d in enumerate(records):

        # Compute the baseline if we're the first record of the pulse,
        # otherwise take the last baseline we've seen in the channel
        if d.record_i == 0:
            bl = last_bl_in[d.channel] = d.data[:baseline_samples].mean()
        else:
            bl = last_bl_in[d.channel]

        # Subtract baseline from all data samples in the record
        # (any additional zeros should be kept at zero)
        last = min(samples_per_record,
                   d.pulse_length - d.record_i * samples_per_record)
        d.data[:last] = int(bl) - d.data[:last]
        d.baseline = bl


@export
@numba.jit(nopython=True, nogil=True, cache=True)
def integrate(records):
    for i, r in enumerate(records):
        records[i]['area'] = r['data'].sum()


@export
@numba.jit(nopython=True, nogil=True, cache=True)
def record_links(records):
    """Return (prev_r, next_r), each arrays of indices of previous/next
    record in the same pulse, or -1 if this is not applicable
    """
    # TODO: needs tests
    if not len(records):
        return
    n_channels = records['channel'].max() + 1
    samples_per_record = len(records[0]['data'])
    previous_record = np.ones(len(records), dtype=np.int32) * NOT_APPLICABLE
    next_record = np.ones(len(records), dtype=np.int32) * NOT_APPLICABLE

    # What was the index of the last record seen in each channel?
    last_record_seen = np.ones(n_channels, dtype=np.int32) * NOT_APPLICABLE
    # What would the start time be of a record that continues that record?
    expected_next_start = np.zeros(n_channels, dtype=np.int64)

    for i, r in enumerate(records):
        ch = r['channel']
        last_i = last_record_seen[ch]
        if r['record_i'] == 0:
            # Record starts a new pulse
            previous_record[i] = NOT_APPLICABLE

        elif r['time'] == expected_next_start[ch]:
            # Continuing record.
            previous_record[i] = last_i
            next_record[last_i] = i

        # (If neither matches, this is a continuing record, but the starting
        #  record has been cut away (e.g. for data reduction))

        last_record_seen[ch] = i
        expected_next_start[ch] = samples_per_record * r['dt']

    return previous_record, next_record


# Chunk size should be at least a thousand,
# else copying buffers / switching context dominates over actual computation
# No max_duration argument: hits terminate at record boundaries, and
# anyone insane enough to try O(sec) long records deserves to be punished
@export
@strax.growing_result(strax.hit_dtype, chunk_size=int(1e4))
@numba.jit(nopython=True, nogil=True, cache=True)
def find_hits(records, threshold=15, _result_buffer=None):
    """Return hits (intervals above threshold) found in records.
    Hits that straddle record boundaries are split (TODO: fix this?)

    NB: returned hits are NOT sorted yet!
    """
    buffer = _result_buffer
    if not len(records):
        return
    samples_per_record = len(records[0]['data'])
    offset = 0

    for record_i, r in enumerate(records):
        # print("Starting record ', record_i)
        in_interval = False
        hit_start = -1

        for i in range(samples_per_record):
            # We can't use enumerate over r['data'], numba gives error
            # TODO: file issue?
            above_threshold = r['data'][i] > threshold
            # print(r['data'][i], above_threshold, in_interval, hit_start)

            if not in_interval and above_threshold:
                # Start of a hit
                in_interval = True
                hit_start = i

            if in_interval:
                if not above_threshold:
                    # Hit ends at the start of this sample
                    hit_end = i
                    in_interval = False

                elif i == samples_per_record - 1:
                    # Hit ends at the *end* of this sample
                    # (because the record ends)
                    hit_end = i + 1
                    in_interval = False

                if not in_interval:
                    # print('saving hit')
                    # Hit is done, add it to the result
                    res = buffer[offset]
                    res['left'] = hit_start
                    res['right'] = hit_end
                    res['time'] = r['time'] + hit_start * r['dt']
                    # Note right bound is exclusive, no + 1 here:
                    res['length'] = hit_end - hit_start
                    res['dt'] = r['dt']
                    res['channel'] = r['channel']
                    res['record_i'] = record_i

                    # Yield buffer to caller if needed
                    offset += 1
                    if offset == len(buffer):
                        yield offset
                        offset = 0

                    # Clear stuff, just for easier debugging
                    # hit_start = 0
                    # hit_end = 0
    yield offset



"""Utilities for dealing with streams of numpy (record) arrays
"""
import itertools

import numpy as np
import strax
export, __all__ = strax.exporter()


@export
class ChunkPacer:

    def __init__(self, source, dtype=None):
        self.dtype = dtype
        self.source = source
        self.buffer = []
        self.buffer_items = 0

    def _fetch_another(self):
        x = next(self.source)
        self.buffer.append(x)
        self.buffer_items += len(x)

    def _squash_buffer(self):
        """Squash the buffer into a single array using np.concatenate"""
        if len(self.buffer) > 1:
            self.buffer = [np.concatenate(self.buffer)]

        # Sanity check on buffer_items
        if not len(self.buffer):
            assert self.buffer_items == 0
        else:
            assert self.buffer_items == len(self.buffer[0])

    def _take_from_buffer(self, n):
        self._squash_buffer()
        if self.buffer_items == 0:
            raise StopIteration
        b = self.buffer[0]

        n = min(n, len(b))
        result, b = np.split(b, [n])
        self.buffer = [b]
        self.buffer_items = len(b)
        return result

    def get_n(self, n: int):
        """Return array of the next n elements produced by source,
        or (if this is less) as many as the source can still produce.
        """
        try:
            while self.buffer_items < n:
                self._fetch_another()
        except StopIteration:
            pass

        return self._take_from_buffer(n)

    def get_until(self, threshold, func=None):
        """Return remaining elements of source below or at threshold,
        assuming source gives sorted arrays.

        :param func: computation to do on array elements before comparison
        """
        if func is None:
            def func(x):
                return x

        if not len(self.buffer):
            self._fetch_another()
            assert len(self.buffer) == 1

        try:
            while (not len(self.buffer[-1])
                    or not func(self.buffer[-1][-1]) > threshold):
                self._fetch_another()
        except StopIteration:
            pass

        n = strax.first_index_not_below(func(self.buffer[-1]), threshold)
        n += sum(len(x) for x in self.buffer[:-1])
        return self._take_from_buffer(n)

    def _put_back_at_start(self, x):
        self.buffer = [x] + self.buffer
        self.buffer_items += 1
        self._squash_buffer()

    def peek(self, n=1):
        x = self.get_n(n)
        self._put_back_at_start(x)
        return x

    @property
    def itemsize(self):
        if self.dtype is None:
            # Peek at one item to figure out the dtype and size
            self.dtype = self.peek().dtype
        return np.zeros(1, dtype=self.dtype).nbytes


@export
def fixed_length_chunks(source, n=10):
    """Yield arrays of maximum length n"""
    p = ChunkPacer(source)
    try:
        while True:
            yield p.get_n(n)
    except StopIteration:
        return


@export
def fixed_size_chunks(source, n_bytes=int(1e8), dtype=None):
    """Yield arrays of maximum size n_bytes"""
    p = ChunkPacer(source, dtype=dtype)
    n = int(n_bytes / p.itemsize)
    try:
        while True:
            yield p.get_n(n)
    except StopIteration:
        return


@export
def alternating_size_chunks(source, *sizes):
    """Yield arrays of sizes[0], then sizes[1], ... sizes[n],
    then sizes[0], etc."""
    p = ChunkPacer(source)
    ns = np.floor(np.array(sizes) / p.itemsize).astype(np.int)
    i = 0
    while True:
        try:
            yield p.get_n(ns[i])
        except StopIteration:
            return
        i = (i + 1) % len(sizes)


@export
def alternating_duration_chunks(source, *durations):
    """Yield arrays of sizes[0], then sizes[1], ... sizes[n],
    then sizes[0], etc."""
    p = ChunkPacer(source)
    t = p.peek()[0]['time']
    i = 0
    while True:
        try:
            t += durations[i]
            yield p.get_until(t, func=strax.endtime)
        except StopIteration:
            return
        i = (i + 1) % len(durations)


@export
def chunk_by_break(source,
                   safe_break,
                   ignore_below=10,
                   max_t_buffer=int(1e10)):
    """Yield arrays whose final elements are separated by at least
    safe_break from the first elements of the next array.
    :param source: Iterator producing interval-like arrays
    :param safe_break: break the LAST time a gap this large occurs in a chunk.
    If no such gap occurs, saves a chunk in the buffer.
    :param ignore_below: Exclude intervals with area lower than this from the
    break calculation
    :param max_buffer: Maximum duration of time in the buffer.
    If a larger buffer would result, break the buffer at the largest gap
    even if it is smaller than safe_break.
    """
    # TODO: needs tests!
    # TODO: add functionality to ChunkPacer instead?
    buffer = None
    for chunk_i, x in enumerate(source):
        if chunk_i == 0:
            buffer = x.copy()
        else:
            buffer = np.concatenate([buffer, x])

        while True:
            large_peaks = buffer[buffer['area'] > ignore_below]
            time_in_buffer = buffer['time'][-1] - buffer['time'][0]
            try:
                break_i = strax.find_break_i(
                    large_peaks,
                    safe_break=safe_break,
                    tolerant=time_in_buffer > max_t_buffer)

                # TODO: can be faster...
                mask = buffer['time'] < large_peaks[break_i]['time']
                yield buffer[mask]
                buffer = buffer[~mask]
            except strax.NoBreakFound:
                break

    if buffer is None:
        print("No data????")

    if buffer is not None and len(buffer):
        yield buffer


@export
def same_length(*sources):
    """Yield tuples of arrays of the same number of items
    """
    pacemaker = sources[0]
    others = [ChunkPacer(s) for s in sources[1:]]

    for x in pacemaker:
        yield tuple([x] + [s.get_n(len(x)) for s in others])


@export
def same_stop(*sources, field=None, func=None):
    """Yield tuples of arrays whose values (of field_name) are below common
    thresholds (set by the chunking of sources[0])
    assumes sources are sorted by field (or return value of func)
    """
    pacemaker = sources[0]
    others = [ChunkPacer(s) for s in sources[1:]]

    for x in pacemaker:
        if not len(x):
            continue
        threshold = x[-1]
        if field is not None:
            threshold = threshold[field]
        if func is not None:
            threshold = func(threshold)
        yield tuple([x] + [s.get_until(threshold, func=func)
                           for s in others])


@export
def sync_iters(chunker, sources):
    """Return dict of iterators over sources (dict name -> iter),
    synchronized using chunker
    """
    names = list(sources.keys())
    sources = list(sources.values())

    teed = itertools.tee(chunker(*sources),
                         len(sources))

    def get_item(iterable, index):
        for x in iterable:
            yield x[index]

    return {names[i]: get_item(teed[i], i)
            for i in range(len(names))}


@export
def merge_iters(iters):
    try:
        while True:
            yield strax.merge_arrs([next(it)
                                    for it in iters])
    except StopIteration:
        return



"""Read/write numpy arrays to/from compressed files or file-like objects
"""
from functools import partial
import bz2

import numpy as np
import blosc
import zstd

import strax
export, __all__ = strax.exporter()

blosc.set_releasegil(True)


COMPRESSORS = dict(
    bz2=dict(
        compress=bz2.compress,
        decompress=bz2.decompress),
    zstd=dict(
        compress=zstd.compress,
        decompress=zstd.decompress),
    blosc=dict(
        compress=partial(blosc.compress, shuffle=False),
        decompress=blosc.decompress),
)


@export
def load_file(f, compressor, dtype):
    """Read and return data from file

    :param f: file name or handle to read from
    :param compressor: compressor to use for decompressing. If not passed,
    will try to load it from json metadata file.
    :param dtype: numpy dtype of data to load
    """
    if isinstance(f, str):
        with open(f, mode='rb') as f:
            return _load_file(f, compressor, dtype)
    else:
        return _load_file(f, compressor, dtype)


def _load_file(f, compressor, dtype):
    data = COMPRESSORS[compressor]['decompress'](f.read())
    return np.frombuffer(data, dtype=dtype)


@export
def save_file(f, data, compressor='zstd'):
    """Save data to file and return number of bytes written

    :param f: file name or handle to save to
    :param data: data (numpy array) to save
    :param compressor: compressor to use
    """
    if isinstance(f, str):
        with open(f, mode='wb') as f:
            return _save_file(f, data, compressor)
    else:
        return _save_file(f, data, compressor)


def _save_file(f, data, compressor='zstd'):
    assert isinstance(data, np.ndarray), "Please pass a numpy array"
    d_comp = COMPRESSORS[compressor]['compress'](data)
    f.write(d_comp)
    return len(d_comp)



"""Functions to perform in-place pulse-level data reduction"""
import numpy as np
import numba
from enum import IntEnum

from strax.processing.pulse_processing import NOT_APPLICABLE, record_links
from strax.processing.peak_building import find_peaks
from .general import fully_contained_in
from strax.dtypes import peak_dtype

__all__ = 'ReductionLevel cut_baseline cut_outside_hits ' \
          'replace_with_spike exclude_tails'.split()


class ReductionLevel(IntEnum):
    """Identifies what type of data reduction has been used on a record
    """
    # Record not modified
    NO_REDUCTION = 0
    # Samples near pulse start/end were removed
    BASELINE_CUT = 1
    # Samples far from a threshold excursion were removed
    HITS_ONLY = 2
    # The record has been replaced with a simpler waveform
    WAVEFORM_REPLACED = 3
    # The raw waveform has been deleted, only metadata survives
    METADATA_ONLY = 4


@numba.jit(nopython=True, nogil=True, cache=True)
def cut_baseline(records, n_before=48, n_after=30):
    """"Replace first n_before and last n_after samples of pulses by 0
    """
    # TODO: records.data.shape[1] gives a numba error (file issue?)
    if not len(records):
        return
    samples_per_record = len(records[0]['data'])

    for d_i, d in enumerate(records):
        if d.record_i == 0:
            d.data[:n_before] = 0

        clear_from = d.pulse_length - n_after
        clear_from -= d.record_i * samples_per_record
        clear_from = max(0, clear_from)
        if clear_from < samples_per_record:
            d.data[clear_from:] = 0

    records.reduction_level[:] = ReductionLevel.BASELINE_CUT


@numba.jit(nopython=True, nogil=True, cache=True)
def cut_outside_hits(records, hits, left_extension=2, right_extension=15):
    """Zero record waveforms not within left_extension or right_extension of
    hits.
    These extensions properly account for breaking of pulses into records.

    If you pass an incomplete (e.g. cut) set of records, we will not save
    data around hits found in the removed records, even if this stretches
    into records that you did pass.
    """
    if not len(records):
        return
    samples_per_record = len(records[0]['data'])

    # For every sample, store if we can cut it or not
    can_cut = np.ones((len(records), samples_per_record), dtype=np.bool_)

    previous_record, next_record = record_links(records)

    for hit_i in range(len(hits)):
        h = hits[hit_i]
        rec_i = h['record_i']

        # Keep required samples in current record
        start_keep = h['left'] - left_extension
        end_keep = h['right'] + right_extension
        can_cut[rec_i][max(0, start_keep):
                       min(end_keep, samples_per_record)] = 0

        # Keep samples in previous/next record if applicable
        if start_keep < 0:
            prev_r = previous_record[rec_i]
            if prev_r != NOT_APPLICABLE:
                can_cut[prev_r][start_keep:] = 0
        if end_keep > samples_per_record:
            next_r = next_record[rec_i]
            if next_r != NOT_APPLICABLE:
                can_cut[next_r][:end_keep - samples_per_record] = 0

    # This is actually quite slow. Perhaps the [:] forces a copy?
    # Without it, however, numba complains...
    for i in range(len(can_cut)):
        records[i]['data'][:] *= ~can_cut[i]
    records.reduction_level[:] = ReductionLevel.HITS_ONLY


@numba.jit(nopython=True, nogil=True, cache=True)
def replace_with_spike(records, also_for_multirecord_pulses=False):
    """Replaces the waveform in each record with a spike of the same integral
    :param also_for_multirecord_pulses: if True, does this even if the pulse
    spans multiple records (so you'll get more than one spike...)
    """
    if not len(records):
        return
    samples_per_record = len(records[0]['data'])

    for i, d in enumerate(records):
        if not (d.record_i == 0 or also_for_multirecord_pulses):
            continue
        # What is the center of this record? It's nontrivial since
        # some records have parts that do not represent data at the end
        center = int(min(d.total_length - samples_per_record * d.record_i,
                         samples_per_record) // 2)
        integral = d.data.sum()
        d.data[:] = 0
        d.data[center] = integral

    records.reduction_level[:] = ReductionLevel.WAVEFORM_REPLACED


# Cannot jit this guy, find_peaks is not a jitted function
def exclude_tails(records, to_pe,
                  min_area=int(2e5),
                  peak_duration=int(1e4),
                  tail_duration=int(1e7),
                  gap_threshold=300):
    """Return records that do not lie fully in tail after a big peak"""
    # Find peaks using the records as "hits". This is rough, but good enough.
    cut = find_peaks(records, to_pe,
                     gap_threshold=gap_threshold,
                     min_area=min_area,
                     result_dtype=peak_dtype(records['channel'].max() + 1),
                     max_duration=peak_duration)
    # Transform these 'peaks' to ranges to cut.
    # We want to cut tails after peaks, not the peaks themselves.
    cut['time'] += peak_duration        # Don't cut the actual peak
    cut['length'] = tail_duration / cut['dt']
    return records[fully_contained_in(records, cut) == -1]



import numpy as np
import numba

from strax import utils
from strax.dtypes import peak_dtype, DIGITAL_SUM_WAVEFORM_CHANNEL

__all__ = 'find_peaks sum_waveform'.split()


@utils.growing_result(dtype=peak_dtype(), chunk_size=int(1e4))
@numba.jit(nopython=True, nogil=True, cache=True)
def find_peaks(hits, to_pe,
               gap_threshold=300,
               left_extension=20, right_extension=150,
               min_hits=3, min_area=0,
               max_duration=int(1e9),
               _result_buffer=None, result_dtype=None):
    """Return peaks made from grouping hits together
    Assumes all hits have the same dt
    :param hits: Hit (or any interval) to group
    :param left_extension: Extend peaks by this many ns left
    :param right_extension: Extend peaks by this many ns right
    :param gap_threshold: No hits for this much ns means new peak
    :param min_hits: Peaks with less than min_hits are not returned
    :param min_area: Peaks with less than min_area are not returned
    :param max_duration: Peaks are forcefully ended after this many ns
    """
    buffer = _result_buffer
    offset = 0
    if not len(hits):
        return
    assert hits[0]['dt'] > 0, "Hit does not indicate sampling time"
    assert min_hits > 0, "min_hits must be > 1"
    assert gap_threshold > left_extension + right_extension, \
        "gap_threshold must be larger than left + right extension"
    assert max_duration / hits[0]['dt'] < np.iinfo(np.int32).max, \
        "Max duration must fit in a 32-bit signed integer"
    # If you write it like below, you get integer wraparound errors
    # TODO :-( File numba issue?
    # assert max_duration < np.iinfo(np.int32).max * hits[0]['dt'], \
    #   "Max duration must fit in a 32-bit signed integer"

    area_per_channel = np.zeros(len(buffer[0]['area_per_channel']),
                                dtype=np.int32)
    in_peak = False
    peak_endtime = 0

    for hit_i, hit in enumerate(hits):
        p = buffer[offset]
        t0 = hit['time']
        dt = hit['dt']
        t1 = hit['time'] + dt * hit['length']

        if not in_peak:
            # This hit starts a new peak candidate
            area_per_channel *= 0
            peak_endtime = t1
            p['time'] = t0 - left_extension
            p['channel'] = DIGITAL_SUM_WAVEFORM_CHANNEL
            p['dt'] = dt
            # These are necessary as prev peak may have been rejected:
            p['n_hits'] = 0
            p['area'] = 0
            in_peak = True

        # Add hit's properties to the current peak candidate
        p['n_hits'] += 1
        peak_endtime = max(peak_endtime, t1)
        p['area_per_channel'][hit['channel']] += hit['area']
        p['area'] += hit['area'] * to_pe[hit['channel']]

        # Look at the next hit to see if THIS hit is the last in a peak.
        # If this is the final hit, it is last by definition.
        if (hit_i == len(hits) - 1
                or hits[hit_i+1]['time'] - peak_endtime >= gap_threshold):
            # Next hit (if it exists) will initialize the new peak candidate
            in_peak = False

            # Do not save if tests are not met. Next hit will erase temp info
            if not (p['n_hits'] >= min_hits and p['area'] >= min_area):
                continue

            # Compute final quantities
            p['length'] = (peak_endtime - p['time'] + right_extension) / dt
            p['area_per_channel'][:] = area_per_channel

            # Save the current peak, advance the buffer
            offset += 1
            if offset == len(buffer):
                yield offset
                offset = 0

    yield offset


@numba.jit(nopython=True, nogil=True, cache=True)
def sum_waveform(peaks, records, adc_to_pe):
    """Compute sum waveforms for all peaks in peaks
    Will downsample sum waveforms if they do not fit in per-peak buffer

    Assumes all peaks and pulses have the same dt
    """
    if not len(records):
        return
    samples_per_record = len(records[0]['data'])
    dt = records[0]['dt']
    time_per_record = samples_per_record * dt
    sum_wv_samples = len(peaks[0]['data'])

    # Big buffer to hold even largest sum waveforms
    # Need a little more even for downsampling..
    swv_buffer = np.zeros(peaks['length'].max() * 2, dtype=np.float32)

    # Index to a window of records
    left_r_i = 0

    for peak_i, p in enumerate(peaks):
        # print("Peak ", peak_i)
        # Clear the relevant part of the swv buffer for use
        # (we clear a bit extra for use in downsampling)
        p_length = p['length']
        swv_buffer[:min(2 * p_length, len(swv_buffer))] = 0

        # Find first record that could contribute to peak
        for left_r_i in range(left_r_i, len(records)):
            if records[left_r_i]['time'] + time_per_record >= p['time']:
                break

        if left_r_i == len(records):
            # No record contributes to the last peak!
            break

        # Scan ahead over records that contribute
        for right_r_i in range(left_r_i, len(records)):
            r = records[right_r_i]
            ch = r['channel']
            # print("record ", right_r_i, ' [', r['time'], ',',
            # r['time'] + r['length'], ')')

            s = int((p['time'] - r['time']) // dt)
            n_r = samples_per_record
            n_p = p_length

            if s < -n_p:
                # print("out of range! s is ", s)
                # Record is fully out of range
                break

            # Range of record that contributes to peak
            r_start = max(0, s)
            r_end = min(n_r, s + n_p)

            # TODO Do we need .astype(np.int32).sum() ??
            p['area_per_channel'][ch] += r['data'][r_start:r_end].sum()

            # Range of peak that receives record
            p_start = max(0, -s)
            p_end = min(n_p, -s + n_r)

            assert p_end - p_start == r_end - r_start, "Ouch, off-by-one error"

            # print("contributes ", r_start, " to ", r_end)
            # print("insert in ", p_start, " to ", p_end)
            if p_end - p_start > 0:
                swv_buffer[p_start:p_end] += \
                    r['data'][r_start:r_end] * adc_to_pe[ch]

        # Store the sum waveform
        # Do we need to downsample the swv to store it?
        downs_f = int(np.ceil(p_length / sum_wv_samples))
        if downs_f > 1:
            # New number of samples in the peak
            new_ns = p['length'] = int(np.ceil(p_length / downs_f))
            p['data'][:new_ns] = \
                swv_buffer[:new_ns * downs_f].reshape(-1, downs_f).sum(axis=1)
            p['dt'] *= downs_f
        else:
            p['data'][:p_length] = swv_buffer[:p_length]

        # Store the total area
        p['area'] = (p['area_per_channel'] * adc_to_pe).sum()



from ast import literal_eval
import logging
import sys
import time
import traceback
import typing

import numpy as np

import strax
export, __all__ = strax.exporter()


@export
class CacheKey(typing.NamedTuple):
    run_id: str
    data_type: str
    lineage: dict

    def __repr__(self):
        return '_'.join([self.run_id,
                         self.data_type,
                         strax.deterministic_hash(self.lineage)])


@export
class NotCached(Exception):
    pass


@export
class NotProvided(Exception):
    pass


class Store:
    """Storage backend for strax data
    """

    saver_init_doc = """\
        :param recover: Load data even if an exception occurred
            during computation (and the data is thus likely incomplete)
        :param take_only: Provide only these data types.
        :param exclude: Do NOT provide these data types.

        If take_only and exclude are both omitted, provide all data types.
        If a data type is listed in both, it will not be provided.

        Attempting to read unwanted data types throws NotCached.
        Attempting to save unwanted data types throws RuntimeError
            (you're supposed to check this with the .provides method).
    """

    def __init__(self,
                 take_only=tuple(), exclude=tuple(),
                 recover=False,
                 readonly=False):
        self._take_only = strax.to_str_tuple(take_only)
        self._exclude = strax.to_str_tuple(exclude)
        self.recover = recover
        self.readonly = readonly
        self.log = logging.getLogger(self.__class__.__name__)

    def provides(self, data_type, write=False):
        """Return whether this store can provides this datatype
        :param write: if True, return if we can also store the data (in case
        we do not have it yet)
        """
        if write and self.readonly:
            self.log.debug(f"Can't save {data_type}, we're readonly")
            return False
        if len(self._exclude) and data_type in self._exclude:
            self.log.debug(f"Not saving {data_type}, excluded")
            return False
        if len(self._take_only) and data_type not in self._take_only:
            self.log.debug(f"Not saving {data_type}, not taken")
            return False
        return True

    def has(self, key: CacheKey):
        try:
            self.find(key)
        except NotCached:
            return False
        return True

    def find(self, key: CacheKey):
        """Return something to load key from (e.g. a directory name)
        or raise NotCached
        """
        something = self._find(key)
        if 'exception' in self._read_meta(something):
            self.log.info(f"Found incomplete data for {key}")
            if self.recover:
                self.log.info("Recovering")
                return something
            self.log.info("Not recovering")
            raise NotCached
        return something

    def loader(self, key: CacheKey, executor=None):
        """Return generator over cached results,
        or raise NotCached if the data is unavailable.
        """
        return self._read(self.find(key), executor)

    def load_meta(self, key: CacheKey):
        return self._read_meta(self.find(key))

    def saver(self, key, metadata):
        if not self.provides(key.data_type):
            raise RuntimeError(f"{key.data_type} not provided by "
                               f"storage backend.")

        self.log.debug(f"Saving {key}")

        metadata.setdefault('compressor', 'blosc')
        metadata['strax_version'] = strax.__version__
        if 'dtype' in metadata:
            metadata['dtype'] = metadata['dtype'].descr.__repr__()
        # >>> Child class should finish saver creation! <<<

    def _read(self, something, executor=None):
        """Iterates over strax results in something (e.g. a directory
        or database collection)
        """
        metadata = self._read_meta(something)
        if not len(metadata['chunks']):
            self.log.warning(f"No actual data in {something}?")
        dtype = literal_eval(metadata['dtype'])
        compressor = metadata['compressor']

        for chunk_info in metadata['chunks']:
            kwargs = dict(chunk_info=chunk_info,
                          dtype=dtype,
                          compressor=compressor)

            if executor is None:
                yield self._read_chunk(something, **kwargs)
            else:
                yield executor.submit(self._read_chunk, something, **kwargs)

    def _find(self, key):
        """Return something self.read reads from, or NotCachedException"""
        raise NotImplementedError

    def _read_meta(self, something):
        raise NotImplementedError

    def _read_chunk(self, something, chunk_info, dtype, compressor):
        raise NotImplementedError


@export             # Needed for type hints elsewhere
class Saver:
    """Interface for saving a data type

    Must work even if forked.
    Do NOT add unpickleable things as attributes (such as loggers)!
    """
    closed = False
    meta_only = False
    prefer_rechunk = True

    def __init__(self, key, metadata):
        self.key = key
        self.md = metadata
        self.md['writing_started'] = time.time()
        self.md['chunks'] = []

    def save_from(self, source: typing.Iterable, rechunk=True):
        """Iterate over source and save the results under key
        along with metadata
        """
        if rechunk and self.prefer_rechunk:
            source = strax.fixed_size_chunks(source)

        try:
            for chunk_i, s in enumerate(source):
                self.save(data=s, chunk_i=chunk_i)
        except strax.MailboxKilled:
            # Write exception (with close), but exit gracefully.
            # One traceback on screen is enough
            self.close()
            pass
        finally:
            if not self.closed:
                self.close()

    def save(self, data: np.ndarray, chunk_i: int):
        if self.closed:
            raise RuntimeError(f"{self.key.data_type} saver already closed!")

        chunk_info = dict(chunk_i=chunk_i,
                          n=len(data),
                          nbytes=data.nbytes)
        if 'time' in data[0].dtype.names:
            for desc, i in (('first', 0), ('last', -1)):
                chunk_info[f'{desc}_time'] = int(data[i]['time'])
                chunk_info[f'{desc}_endtime'] = int(strax.endtime(data[i]))

        if not self.meta_only:
            chunk_info.update(self._save_chunk(data, chunk_info))
        self._save_chunk_metadata(chunk_info)

    def _save_chunk(self, data, chunk_info):
        raise NotImplementedError

    def _save_chunk_metadata(self, chunk_info):
        raise NotImplementedError

    def close(self):
        if self.closed:
            raise RuntimeError(f"{self.key.data_type} saver already closed!")
        self.closed = True
        exc_info = sys.exc_info()
        if exc_info[0] not in [None, StopIteration]:
            self.md['exception'] = traceback.format_exc()
        self.md['writing_ended'] = time.time()

        # >>> child class may want to finish writing metadata <<<



"""Fundamental dtypes for use in strax.

Note that if you change the dtype titles (comments), numba will crash if
there is an existing numba cache. Clear __pycache__ and restart.
TODO: file numba issue.
"""
import numpy as np

__all__ = ('interval_dtype record_dtype hit_dtype peak_dtype '
           'DIGITAL_SUM_WAVEFORM_CHANNEL DEFAULT_RECORD_LENGTH').split()

DIGITAL_SUM_WAVEFORM_CHANNEL = -1
DEFAULT_RECORD_LENGTH = 110


# Base dtype for interval-like objects (pulse, peak, hit)
interval_dtype = [
    (('Channel/PMT number',
        'channel'), np.int16),
    (('Time resolution in ns',
        'dt'), np.int16),
    (('Start time of the interval (ns since unix epoch)',
        'time'), np.int64),
    # Don't try to make O(second) long intervals!
    (('Length of the interval in samples',
        'length'), np.int32),
    # Sub-dtypes MUST contain an area field
    # However, the type varies: float for sum waveforms (area in PE)
    # and int32 for per-channel waveforms (area in ADC x samples)
]


def record_dtype(samples_per_record=DEFAULT_RECORD_LENGTH):
    """Data type for a waveform record.

    Length can be shorter than the number of samples in data,
    this indicates a record with zero-padding at the end.
    """
    return interval_dtype + [
        (("Integral in ADC x samples",
            'area'), np.int32),
        # np.int16 is not enough for some PMT flashes...
        (('Length of pulse to which the record belongs (without zero-padding)',
            'pulse_length'), np.int32),
        (('Fragment number in the pulse',
            'record_i'), np.int16),
        (('Baseline in ADC counts. data = int(baseline) - data_orig',
            'baseline'), np.float32),
        (('Level of data reduction applied (strax.ReductionLevel enum)',
            'reduction_level'), np.uint8),
        # Note this is defined as a SIGNED integer, so we can
        # still represent negative values after subtracting baselines
        (('Waveform data in ADC counts above baseline',
            'data'), np.int16, samples_per_record),
    ]


# Data type for a 'hit': a sub-range of a record
hit_dtype = interval_dtype + [
    (("Integral in ADC x samples",
        'area'), np.int32),
    (('Index of sample in record in which hit starts',
        'left'), np.int16),
    (('Index of first sample in record just beyond hit (exclusive bound)',
        'right'), np.int16),
    (('Internal (temporary) index of fragment in which hit was found',
        'record_i'), np.int32),
]


def peak_dtype(n_channels=100, n_sum_wv_samples=200, n_widths=11):
    """Data type for peaks - ranges across all channels in a detector
    Remember to set channel to -1 (todo: make enum)
    """
    if n_channels == 1:
        raise ValueError("Must have more than one channel")
        # Otherwise array changes shape?? badness ensues
    return interval_dtype + [
        (('Integral across channels in photoelectrons',
            'area'), np.float32),
        # Area per channel in ADC * samples
        (('Integral per channel in ADX x samples (not PE!)',
            'area_per_channel'), np.int32, n_channels),
        # Number of hits from which this peak was constructed
        # (zero if peak was split afterwards)
        (("Number of hits from which peak was constructed "
          "(currently zero if peak is split afterwards)",
            'n_hits'), np.int16),
        # Waveform data in PE/sample
        (('Waveform data in PE/sample (not PE/ns!)',
            'data'), np.float32, n_sum_wv_samples),
        (('Peak widths in ns: range of central area fraction',
            'width'), np.float32, n_widths)
    ]



import numpy as np
import numba
import strax

__all__ = 'split_peaks '.split()


def split_peaks(peaks, records, to_pe, min_height=25, min_ratio=4):
    """Return peaks after splitting at prominent sum waveform minima
    'Prominent' means: on either side of a split point, local maxima are:
      - larger than minimum + min_height
      - larger than minimum * min_ratio
    (this is related to topographical prominence for mountains)

    Min_height is in pe/ns (NOT pe/bin!)
    """
    is_split = np.zeros(len(peaks), dtype=np.bool_)

    new_peaks = _split_peaks(peaks,
                             min_height=min_height,
                             min_ratio=min_ratio,
                             orig_dt=records[0]['dt'],
                             is_split=is_split,
                             result_dtype=peaks.dtype)
    strax.sum_waveform(new_peaks, records, to_pe)
    return strax.sort_by_time(np.concatenate([peaks[~is_split],
                                              new_peaks]))


@strax.utils.growing_result(dtype=strax.peak_dtype(), chunk_size=int(1e4))
@numba.jit(nopython=True, nogil=True, cache=True)
def _split_peaks(peaks, min_height, min_ratio, orig_dt, is_split,
                 _result_buffer=None, result_dtype=None):
    new_peaks = _result_buffer
    offset = 0

    for p_i, p in enumerate(peaks):
        prev_split_i = 0

        for split_i in find_split_points(p['data'][:p['length']],
                                         min_height=min_height * p['dt'],
                                         min_ratio=min_ratio):
            is_split[p_i] = True

            r = new_peaks[offset]
            r['time'] = p['time'] + prev_split_i * p['dt']
            r['channel'] = p['channel']
            r['dt'] = orig_dt
            r['length'] = (split_i - prev_split_i) * p['dt'] / orig_dt

            offset += 1
            if offset == len(new_peaks):
                yield offset
                offset = 0

            prev_split_i = split_i

    yield offset


@numba.jit(nopython=True, nogil=True, cache=True)
def find_split_points(w, min_height=0, min_ratio=0):
    """"Yield indices of prominent local minima in w
    If there was at least one index, yields len(w)-1 at the end
    """
    found_one = False
    last_max = 0
    min_since_max = 99999999999999.9
    min_since_max_i = 0

    for i, x in enumerate(w):
        if x < min_since_max:
            # New minimum since last max
            min_since_max = x
            min_since_max_i = i

        if min(last_max, x) > max(min_since_max + min_height,
                                  min_since_max * min_ratio):
            # Significant local minimum: tell caller,
            # reset both max and min finder
            yield min_since_max_i
            found_one = True
            last_max = x
            min_since_max = 99999999999999.9
            min_since_max_i = i

        if x > last_max:
            # New max, reset minimum finder state
            # Notice this is AFTER the split check,
            # to accomodate very fast rising second peaks
            last_max = x
            min_since_max = 99999999999999.9
            min_since_max_i = i

    if found_one:
        yield len(w) - 1



import os
import json
import zipfile

import strax
from .files import FileStore
from .common import NotCached

export, __all__ = strax.exporter()


@export
class ZipFileStore(FileStore):
    """ZipFile-based storage backend for strax.
    All data for one run is assumed to be in a single zip file <run_id>.zip,
    with the same file/directory structure as created by FileStore.

    This cannot modify the zip files (would result in concurrency hell).
    # TODO: make method to do create zip file more conveniently
    """
    def __init__(self, *args, **kwargs):
        """:param data_dirs: List of data directories to use.
        Entries are preferred for use from first to last.

        {provides_doc}
        """
        super().__init__(*args, **kwargs)
        self.readonly = True

    @staticmethod
    def _get_fn(key):
        return key.data_type + '_' + strax.deterministic_hash(key.lineage)

    def _find(self, key):
        dirn = str(key)

        for zipn in self._candidate_zips(key):
            with zipfile.ZipFile(zipn) as zp:
                try:
                    zp.getinfo(dirn + '/metadata.json')
                except KeyError:
                    continue
                else:
                    self.log.debug(f"{key} is in cache.")
                    return zipn, dirn

        self.log.debug(f"{key} is NOT in cache.")
        raise NotCached

    def _candidate_zips(self, key):
        """Return zipfiles at which data meant by key
        could be found or saved"""
        return [
            os.path.join(d, key.run_id + '.zip')
            for d in self.data_dirs]

    def _read_chunk(self, zipn_and_dirn, chunk_info, dtype, compressor):
        zipn, dirn = zipn_and_dirn
        with zipfile.ZipFile(zipn) as zp:
            with zp.open(dirn + '/' + chunk_info['filename']) as f:
                return strax.load_file(f, dtype=dtype, compressor=compressor)

    def _read_meta(self, zipn_and_dirn):
        zipn, dirn = zipn_and_dirn
        with zipfile.ZipFile(zipn) as zp:
            with zp.open(dirn + '/metadata.json') as f:
                return json.loads(f.read())

    def saver(self, key, metadata):
        raise NotImplementedError



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# Get requirements from requirements.txt, stripping the version tags
with open('requirements.txt') as f:
    requires = [x.strip().split('=')[0]
                for x in f.readlines()]

with open('README.md') as file:
    long_description = file.read()

setup(name='strax',
      version='0.1.1',
      description='Streaming analysis for XENON',
      author='Jelle Aalbers',
      author_email='j.aalbers@uva.nl',
      url='https://github.com/jelleaalbers/strax',
      setup_requires=['pytest-runner'],
      install_requires=requires,
      tests_require=requires + ['pytest',
                                'boltons',
                                'hypothesis'],
      long_description=long_description,
      long_description_content_type="text/markdown",
      packages=['strax',
                'strax.processing',
                'strax.storage',
                'strax.xenon'],
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',],
      zip_safe = False,
    )
